/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*******************************************************************************
  program name: lm_pval.c

  Program to do posterior IBD probs
  The state (X) consists of segregation indicators

  X determines the IBD pattern among the specified descendant gametes.

  The data, Y, is the multilocus genotypes of observed individuals in the
  pedigree.

  lm_pval implemented, based on Feb 2003 Autozyg program lm_auto.c
                                                                EAT Apr 2003
  Modified to use globals identifying affected individuals,
  component info.                                               MAJ Jun 2003

  The realized probability distributions are scores according to one
     of several ibd measures and a p-value of a test for linkage is
     calculated at each marker locus, and a multiple-testing corrected
     one based on the maximum over loci.

  Modified to do correct fuzzy-pval dsns (per Geyer & Thompson) EAT Oct 2004
          First get null-null dsn quantiles accurate to .001 (0.1%)
          Then score marker loci, as generated, against this dsn.
  Generalized to use any ibd measure (currently max_fgl, spair, robdom)
    and to print out quantiles of the fuzzy-p distribution      EAT July 2005

  This version replaces the obsolete lm_pval of MORGAN 2.5 & 2.6
    for MORGAN 2.7

*******************************************************************************/

/**
 * @defgroup lm_pval lm_pval
 * @ingroup autozyg
 * @{
 */

#define PNAM  "lm_pval"

#include "autozyg.h"

int main (int argc, char *argv[])
{
   FILE *oSeedF;

   Measure scoraff;
   MSD_PTR p_msd;

   double **hdsn;                       /* [npos+1][nitnull]
                                         * list of null imputation scores
                                         * at all evaluation positions.       */
   double **fzpval;                     /* [npos+1][nnull]
                                         * summed over component [comp_no].   */

   double **qndsn;                      /* quantiles of the null distribution
                                         * sampled  [npos+1][nnull]
                                         * F^{-1}((k+1)/nnull) is element k.  */
   double *mean_stat;                   /* [npos+1] vector of trad stats.     */

   double *unused = NULL;
   double **prob;

   int   comp_ndx;                      /* Current component subscript.       */
   int   comp_no;                       /* Current component number.          */

   int   nnull = 1000;                  /* Number of quantiles for null.      */
   int   id0;                           /* Indicator for id null marginals.   */
   int   decr;                          /* Number of realizations per nnull.  */

   int   scored_iterations;             /* Number of scored iterations.       */

   int   nitnull, ntpos;
   int   iteration;
   int   nsw;
   int   p1;

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   /* Number of ibd score positions -- currently score only at markers.       */
   ntpos = NMarks;
   /* Number of null realizations -- currently set same as MCMC iterations.   */
   nitnull = NIt;
   /* Indicator of id null marginals -- will always be 1 in this program.     */
   id0 = TRUE;

   /* Set up ibd measure: any of these 3 should currently work.               */

   if (MaxFGLOn)
      scoraff = max_fgl;                /* Default: Used in current Gold.     */
   else if (RobdomOn)
      scoraff = robdom;
   else if (SpairsOn)
      scoraff = spair;

   setup_ibdmeas (scoraff);

   mean_stat = alloc_dbl1d (ntpos + 1);

   /* First generate the null-null distribution quantiles,      EAT  Oct 2004 */
   /*   This will be fixed, and any future sampling of null locus irrelevant. */
   /*    The 0 slot of the hdsn will be used to store the max over loci.      */
   /*   For convenience, NIt iterations are used, but iid not mcmc.           */

   /* Now realize segs and set hdsn from nitnull nulls.                       */
   hdsn = alloc_dbl2d (ntpos + 1, nitnull);

   for (iteration = 0; iteration < nitnull; ++iteration)
   {
      /* Realize the initial switches.                                        */

      realise (NMarks, xecomo, nsw);

      /* Set founder genome labels (gg0, gg1) from switches.                  */

      for (p1 = 1; p1 <= NMarks; ++p1)
         set_FGL (p1);

      /* Score the ibd measure -- store the value.                            */
      multi_gen_score (iteration, scoraff, ntpos, hdsn);
   }

   /* Convert distribution to 0.1 % quantiles.                                */
   qndsn = alloc_dbl2d (ntpos + 1, nnull + 1);
   decr = nitnull / nnull;
   for (p1 = 0; p1 <= ntpos; ++p1)
      make_quantiles (nnull, decr, hdsn[p1], qndsn[p1]);

   /* Free the full set of null distribution realizations: only the quantiles */
   /* are needed.                                                             */
   free_2d ((void *) &hdsn);

   /* Now print quantiles for check -- at some large interval.                */
   printf ("\n Now print null quantiles:\n\n");
   pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
   fflush (stdout);

   /* If marginals should be ident distributed (id0=1), then merge quantiles. */
   if (id0 && (NMarks > 1))
   {
      compress_marginals (1, NMarks, nnull, qndsn);

      /* Now print quantiles again for check -- at some large interval.       */
      printf (" Now print merged null quantiles:\n\n");
      pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
      fflush (stdout);
   }

   /* Allocate and zero the fuzzy-pvals.                                      */
   fzpval = alloc_dbl2d (ntpos + 1, nnull + 1);

   /* Now intialize the inheritance vectors.                                  */
   prob = setupInherit (1);

   p_msd = alloc_msd ();
   alloc_logprob ();

   prntStatusBMC ();                    /* Print status for beginning MCMC.   */

   /* ................  BURN-IN  .......................................      */

   /* Now do burn-in iterations.                                              */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   fflush (stdout);

   for (iteration = 1; iteration <= NBrnIt; ++iteration)
   {
      if ((iteration % ChkIt) == 0)
      {
         printf (" Starting burn-in iteration %d\n", iteration);
         fflush (stdout);
      }
      sampling (iteration, 1, NMarks, nsw, nFGL, prob);
   }
   printf ("\n");
   fflush (stdout);

   /* ................  COMPUTE USING MCMC  ............................      */

   scored_iterations = 0;               /*Initialize scored iterations count. */

   /* Now do the regular MCMC sampling, and score the results.                */

   printf (" Now starting %d iterations of MCMC\n", NIt);
   fflush (stdout);

   for (iteration = 1; iteration <= NIt; ++iteration)
   {
      if (DebugMain)
      {
         fprintf (stderr, "\nAt iteration %d,\n", iteration);
         dbgXXGG ();
      }

      /* Update xx1, xx0.                                                     */
      sampling ((iteration - 1), 1, NMarks, nsw, nFGL, prob);

      if ((iteration % ScorIt) == 0)
      {
         /* Now scoring will be as a fuzzy-p dsn for each locus and the max.  */
         /* If t(S^{(j)}) is exceeded by exactly k of the t(S_0^{(i)}) and    */
         /* tied with w, then the contribution to the histogram is rectangle  */
         /* with area 1/n, and base endpoints k/(m+1), (k+1+w)/(m+1).         */

         score_all (scoraff, fzpval, NMarks, nnull, qndsn, mean_stat);

         update_counts ("", p_msd, 0, 1, prob, unused);
         ++scored_iterations;
      }
   }

   /* Print status that we have finished with the MCMC.                       */
   printf ("\n\n");
   prntStatusFMC ();
   fflush (stdout);

   printf (" Values of expectation statistics, max and by locus:\n");
   for (p1 = 0; p1 <= NMarks; ++p1)
      printf (" %7.3f", mean_stat[p1] / scored_iterations);

   for (p1 = 0; p1 <= NMarks; ++p1)
      norcum (NIt, nnull, fzpval[p1]);
   printf ("\n\n");

   printf (" Combined distribution of fuzzy p-values, by locus:\n");
   printf ("pval maxim");
   pr_pvals (NMarks, fzpval, nnull, nnull / 10);
   fflush (stdout);

   for (comp_ndx = 0; comp_ndx < n_compnts; ++comp_ndx)
   {
      comp_no = comp_ndx + 1;
      if (cmplist[comp_no].naff == 0)
         printf (" There are no affected individuals for component %d\n\n",
                 comp_no);
      else if (PrintLevel >= 4)
      {
         if (cmplist[comp_no].nmeioses == 0)
            printf (" Component %d has no informative meioses\n\n", comp_no);
         else
         {
            pr_cmpnthead (comp_ndx);
            pr_comp_mcmc_stats (p_msd, 1, scored_iterations, comp_no, comp_ndx);
         }
      }
      if ((n_compnts - comp_ndx) > 1)
         printf ("\n");
      fflush (stdout);
   }

   /* Free up allocated memory.                                               */

   free_2d ((void *) &fzpval);
   free_2d ((void *) &qndsn);
   free_1d ((void *) &mean_stat);

   free_msd (p_msd);

   free_logprob ();
   free_xxgg ();
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (NMarks + 1, &prob);
   free_msmpl ();
   free_traits ();
   free_cmplists ();
   free_ibd_measures ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (Scan)
      free_perm ();

   exit (EXIT_SUCCESS);
}

/** @} */
