/******************************************************************************/
/*                                                                            */
/* The beaglesim subroutines:                                                 */
/*                                                                            */
/*     kiss(), kissunif(), initstream(), addmarker(), addedge(), and          */
/*     free_marker_list().                                                    */
/*                                                                            */
/******************************************************************************/

#define PNAM "ibd_create"
#define BUFLEN 256

#include "parsie.h"
#include "rans.h"
#include "ibd_create.h"

/* Random number generation:
 * Algorithm taken from usenet post by (reputable guy) George Marsaglia.
 * "Random numbers for C: The END?"
 * https://groups.google.com/group/sci.math.num-analysis/msg/eb4ddde782b17051?hl=en&pli=1
 *
 * Independent RNG streams are created with initstream (seed).                */

unsigned int kiss (rngstate * st)
{
   st->z = 36969 * (st->z & 65535) + (st->z >> 16);
   st->w = 18000 * (st->w & 65535) + (st->w >> 16);
   st->jsr ^= (st->jsr << 17);
   st->jsr ^= (st->jsr >> 13);
   st->jsr ^= (st->jsr << 5);
   st->jcong = 69069 * st->jcong + 1234567;

   return (((st->z << 16) + st->w) ^ (st->jcong)) + st->jsr;
}

double kissunif (rngstate * st)
{
   return kiss (st) / (double) 4294967296;
}

rngstate *initstream (unsigned int seed)
{
   rngstate *st;
   int   i;

   st = malloc (sizeof (rngstate));
   alloc_chk (WHERE, st);

   st->z = st->w = st->jsr = st->jcong = seed;
   st->z = kiss (st);
   st->w = kiss (st);
   st->jsr = kiss (st);
   st->jcong = kiss (st);

   /* warm up the stream */
   for (i = 0; i < 256; ++i)
      kiss (st);

   return (st);
}

marker *addmarker (marker * currmarker, int index)
{
   marker *newmarker;

   newmarker = malloc (sizeof (marker));
   alloc_chk (WHERE, newmarker);

   currmarker->nextmarker = newmarker;
   newmarker->index = index;
   newmarker->nextmarker = NULL;

   return newmarker;
}

edge *addedge (edge * head, int parent, int child, char allele, int count)
{
   edge *newedge;

   newedge = malloc (sizeof (edge));
   alloc_chk (WHERE, newedge);

   newedge->nextedge = head;
   newedge->parent = parent;
   newedge->child = child;
   newedge->allele = allele;
   newedge->count = count;

   return newedge;
}

/******************************************************************************/
/*                                                                            */
/* The free_marker_list subroutine is called to free up the memory which was  */
/* allocated for the linked list of markers, along with their linked lists of */
/* edges.                                                                     */
/*                                                                            */
/******************************************************************************/

void free_marker_list (marker * first_marker)
{
   marker *curr_marker;
   marker *next_marker;

   edge *curr_edge;
   edge *next_edge;

   curr_marker = first_marker;
   while (curr_marker)
   {
      curr_edge = curr_marker->edges;
      while (curr_edge)
      {
         next_edge = curr_edge->nextedge;
         free (curr_edge);
         curr_edge = next_edge;
      }

      next_marker = curr_marker->nextmarker;
      free (curr_marker);
      curr_marker = next_marker;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The beaglesim subprogram.                                                  */
/*                                                                            */
/******************************************************************************/

int beaglesim (int argc, char **argv)
{
   FILE *dagfile;                       /* Input dag's file control block.    */
   FILE *oScorF;                        /* The beaglesim output results file. */

   rngstate *recombstream;
   rngstate *transstream;

   marker *mhead;
   marker *mcurr;
   edge *ecurr, *ecurr2;
   edge *ehead;

   unsigned int seed1, seed2;

   char  tmarker[BUFLEN];
   char  buf[BUFLEN];
   char  tallele;
   char  c;

   char **outmat;
   char *dagfilename;
   char *outfilename;

   int   numb_out_haps;                 /* Number of haplotypes to generate.  */

   int   currstate, transpool, runcount, transrand, recombflag;
   int   tlevel, tparent, tchild, tcount;
   int   nhaps, currnhaps, nmarks, ninds;
   int   h1, h2, i1, m1;

/******************************************************************************/
/*                                                                            */
/* The executable portion of the main program starts here.                    */
/*                                                                            */
/******************************************************************************/

   /* ................................  SET-UP   ............................ */

   setupBasic ("R", argc, argv, PNAM);

   /* Get the file name of the input extra file.                              */

   dagfilename = get_filent (XTRA)->fname;
   if (dagfilename == NULL)
      give_up (ERD_ixf);

   /* Get the file name of the output scores file.                            */

   outfilename = get_filent (OSCORE)->fname;
   if (outfilename == NULL)
      give_up (ERD_nof);

   printf ("\n");
   fflush (stdout);

#ifdef DEBUG
   testrng ();
#endif

   /* Unzip the input extra file if it is gzipped.                            */

   if (gunzipit (dagfilename))
      give_up (ERD_guz);

   /* Open the input extra file for reading the beaglesim input dagfile.      */

   if (!(dagfile = open_fil (XTRA)))
      give_up (ERR_qit);

   printf ("\n");
   fflush (stdout);

   /* Open the output scores file to use as beaglesim's output results file.  */

   if (!(oScorF = open_fil (OSCORE)))
      give_up (ERR_qit);

   printf ("\n");
   fflush (stdout);

   /* When the user specifies that haplotype labels are to be output, this    */
   /* implies that spaces must also be output.                                */

   if (OutHapsLabs)
      OutWithSpac = TRUE;

   mhead = malloc (sizeof (marker));
   alloc_chk (WHERE, mhead);

   mhead->index = 0;
   mcurr = mhead;

   ehead = NULL;

   fgets (buf, BUFLEN, dagfile);        /* Skip header.                       */
   while (fgets (buf, BUFLEN, dagfile))
   {
      if (buf[0] == '\n')
         continue;                      /* Empty line.                        */

      /* Did we finish the line?  If not, scan ahead to the next line.        */
      for (i1 = 0; i1 < BUFLEN; ++i1)
         if (buf[i1] == '\n')
            break;

      if (BUFLEN == i1)
         while ((c = getc (dagfile), '\n' != c && EOF != c));
      sscanf (buf, "%d %s %d %d %c %d", &tlevel, tmarker, &tparent, &tchild,
              &tallele, &tcount);

      if (tlevel != mcurr->index)
         /* we've moved to the next marker so wrap up the current one.        */
      {
         mcurr->edges = ehead;          /* connect the chain of edges to the marker node */
         ehead = NULL;                  /* start a new edge chain for the next marker */

         /* count total haps at marker and check that it's same as other markers.
          * also calculate total count of edges which share each edge's child.
          */
         currnhaps = 0;
         for (ecurr = mcurr->edges; ecurr; ecurr = ecurr->nextedge)
         {
            currnhaps += (ecurr->count);
            ecurr->nsamechild = 0;
            /* loop over edges and add up counts of edges with same child as ecurr.
             * not the most efficient but easy to write */
            for (ecurr2 = mcurr->edges; ecurr2; ecurr2 = ecurr2->nextedge)
            {
               if (ecurr->child == ecurr2->child)
                  ecurr->nsamechild += ecurr2->count;
            }
         }
         if (mcurr->index == 0)
         {
            nhaps = currnhaps;
            printf (" DAG file contains %d haplotypes.\n", nhaps);
         }
         else
         {
            if (nhaps != currnhaps)
            {
               printf (" Error: marker %d contains %d haplotypes but marker 0\
contains %d haplotypes.  Check DAG file for problems.\n", mcurr->index, currnhaps, nhaps);
               return (5);
            }
         }

         mcurr = addmarker (mcurr, tlevel);
      }
      /* add a new edge node for this edge */
      ehead = addedge (ehead, tparent, tchild, tallele, tcount);
   }
   mcurr->edges = ehead;
   nmarks = mcurr->index + 1;
   fclose (dagfile);

   if (DebugLevel == 2)
   {
      printf ("\n\n");
      disp_dag_file (mhead);
      printf ("\n\n");
      fflush (stdout);
   }

   printf (" Finished reading DAG file.\n\n\n");
   fflush (stdout);

   /* Now do the actual haplotype simulation.                                 */

   getsd (&seed1, &seed2);
   recombstream = initstream (seed1);
   transstream = initstream (seed2);

   if (OUT_GENOTYPES)
   {
      /* There is a mistake in the out_genotypes subroutine of the parser.    */
      /* It saves the number included in an "output ... genotypes parameter   */
      /* statement in the NumbOutHaps global variable.  I believe I need to   */
      /* add a global variable to save the number of output genotypes in.     */
      /* But for now I can just use NumbOutHaps as the number of genotypes.   */

      numb_out_haps = NumbOutHaps * 2;  /* Number of haplotypes to generate.  */
      ninds = NumbOutHaps;
   }
   else
      numb_out_haps = NumbOutHaps;

   outmat = malloc (numb_out_haps * sizeof (char *));
   alloc_chk (WHERE, outmat);

   printf (" Generating %d haplotypes.\n", numb_out_haps);
   for (h1 = 0; h1 < numb_out_haps; ++h1)
   {
      /* Simulation works like this:
       * Say there are two edges with child node 1
       * with counts 10 and 20 at marker k
       * and three edges with parent node 1
       * with counts 5 14 and 11 * at marker k+1.
       *
       * At marker k we chose an edge with child 1, so
       * transpool = 30 (total pool of edges we can transition to).
       * we choose a random integer from 1 to 30, say transrand=12.
       * we traverse all edges at marker k+1, skipping those
       * with parent != 1.  the first edge has count 5,
       * so the running count=runcount
       * is 5, smaller than 12.  the next edge has count 14, so the running
       * count is 19, >= to transrand, we we choose that edge.
       *
       * For a recombinant jump, we randomly choose an edge, so we set
       * transpool = total # of haps and do not skip any edges.
       *
       * Note that edges do not have to be in any kind of order, which
       * simplifies things
       *
       * Two random streams are used so changes can be made to, say,
       * transition process but the recombination patterns stay the same.
       *
       */
      outmat[h1] = malloc (nmarks * sizeof (char));
      alloc_chk (WHERE, outmat[h1]);

      currstate = 0;                    /* Start in state 0 always.           */
      transpool = nhaps;                /* First set of edges all have same parent (0). */

      /* Step through the markers.                                            */
      for (mcurr = mhead; mcurr; mcurr = mcurr->nextmarker)
      {
         /* Do we recombine?  If so, set transpool to all haps.               */
         recombflag = kissunif (recombstream) < LDRelaxProb;
         if (recombflag)
            transpool = nhaps;

         /* choose a random haplotype index out of the pool
          * of possible haplotypes we can transition to */
         transrand = ceil (transpool * kissunif (transstream));
#ifdef DEBUG
         printf ("\nMarker %d\n", mcurr->index);
         printf ("transpool is size %d\n", transpool);
         printf ("Recombine? %s\n", recombflag ? "Yes" : "No");
         if (recombflag)
         {
            printf ("getting hap # %d \n", transrand);
         }
         else
         {
            printf ("getting hap # %d with parent %d\n", transrand, currstate);
         }
#endif

         /* step through edges until we find that haplotype */
         runcount = 0;
         for (ecurr = mcurr->edges; ecurr; ecurr = ecurr->nextedge)
         {
            /* if recombinant, we are choosing a random edge
             * so doesn't matter what the parent is */
#ifdef DEBUG
            printf
               ("This edge has parent %d and child %d and count %d and allele %c \n",
                ecurr->parent, ecurr->child, ecurr->count, ecurr->allele);
            if (!recombflag && ecurr->parent != currstate)
               printf ("Runcount is %d\n", runcount);
#endif
            if (!recombflag && ecurr->parent != currstate)
               continue;
            runcount += ecurr->count;
#ifdef DEBUG
            printf ("Runcount is %d\n", runcount);
            if (transrand <= runcount)
               printf
                  ("we found the hap, new child/parent is %d, allele is %c \n",
                   ecurr->child, ecurr->allele);
#endif
            if (transrand <= runcount)
               break;                   /* found it */
         }
#ifdef DEBUG
         printf ("Press ENTER to continue\n");
         getchar ();
#endif
         currstate = ecurr->child;
         /* at the next marker, size of pool of possible transitions is
          * the # of haps with this node as the parent */
         transpool = ecurr->nsamechild;
         outmat[h1][mcurr->index] = ecurr->allele;
      }
#ifdef DEBUG
      progbar (ceil (100 * (h1 + 1) / (double) numb_out_haps));
#endif
   }
   printf ("\n\n");
   fflush (stdout);

   free (recombstream);
   free (transstream);

   /* Finally output the results to oScorF along with an informative message  */
   /* as to where we are in the program to stdout.                            */

   if (OUT_GENOTYPES)
   {
      printf (" Writing %d genotypes to \"%s\".\n", ninds, outfilename);
      fprintf (oScorF, "set markers %d data\n", nmarks);

      for (i1 = 0; i1 < ninds; ++i1)
      {
         fprintf (oScorF, " indv_%03d", i1 + 1);

         h1 = i1 * 2;                   /* Compute subscripts to haplotypes.  */
         h2 = h1 + 1;

         for (m1 = 0; m1 < nmarks; ++m1)
         {
            fprintf (oScorF, "  %c", outmat[h1][m1]);
            fprintf (oScorF, " %c", outmat[h2][m1]);
         }
#ifdef DEBUG
         progbar (ceil (100 * h2 / (double) NumbOutHaps));
#endif
         fprintf (oScorF, "\n");
      }
   }
   else if (OUT_HAPS_AS_ROWS)
   {
      printf (" Writing %d haplotypes to \"%s\".\n", NumbOutHaps, outfilename);

      for (h1 = 0; h1 < NumbOutHaps; ++h1)
      {
         h2 = h1 + 1;
         if (OutHapsLabs)
            fprintf (oScorF, "%d ", h2);

         for (m1 = 0; m1 < nmarks; ++m1)
         {
            if (OutWithSpac)
               fprintf (oScorF, " ");
            fprintf (oScorF, "%c", outmat[h1][m1]);
         }
#ifdef DEBUG
         progbar (ceil (100 * h2 / (double) NumbOutHaps));
#endif
         fprintf (oScorF, "\n");
      }
   }
   else if (OUT_HAPS_AS_COLS)
   {
      printf (" Writing %d haplotypes to \"%s\".\n", NumbOutHaps, outfilename);

      if (OutHapsLabs)
      {
         for (h2 = 1; h2 <= NumbOutHaps; ++h2)
            fprintf (oScorF, " %d", h2);
         fprintf (oScorF, "\n");
      }

      for (m1 = 0; m1 < nmarks; ++m1)
      {
         for (h1 = 0; h1 < NumbOutHaps; ++h1)
         {
            if (OutWithSpac)
               fprintf (oScorF, " ");
            fprintf (oScorF, "%c", outmat[h1][m1]);
         }
#ifdef DEBUG
         progbar (ceil (100 * (m1 + 1) / (double) nmarks));
#endif
         fprintf (oScorF, "\n");
      }
   }
   printf ("\n");

   fflush (oScorF);
   fclose (oScorF);

#if 0
   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
#endif

   /* We are done, so free up the storage that has been allocated by Morgan   */
   /* to support the beaglesim subroutine.                                    */

   free_marek ();
   free_locnams ();
   free_fnams ();

   /* Free the remaining beaglesim subroutine allocated memory.               */

   free_marker_list (mhead);

   for (h1 = 0; h1 < numb_out_haps; ++h1)
      free (outmat[h1]);
   free (outmat);

   fflush (stdout);

   return 0;
}
