/******************************************************************************/
/*                                                                            */
/* Subroutines used by the ibd_create subprograms.                            */
/*                                                                            */
/******************************************************************************/

#include "markers.h"

#include "ibd_haplo_directory.h"
#include "ibd_create.h"

/******************************************************************************/
/*                                                                            */
/* The fill_hashtable subroutine is called to set the contents of the FGL's   */
/* hashtable passed as the first argument to this subroutine.                 */
/*                                                                            */
/* During the process of doing this the subroutine also needs to count how    */
/* many unique FGL's at any of the selected markers were found in the input   */
/* ibd_graph gamete data provided by the user.                                */
/*                                                                            */
/* The final count of unique FGL's found is returned in the last argument to  */
/* this subroutine.                                                           */
/*                                                                            */
/******************************************************************************/

void fill_hashtable (Hashnode ** hashtable, int nsnps, int *num_uniq_fgls)
{
   Chromseglist *curr_seglist;          /* Current gamete segment list.       */

   int   nextindex = 0;                 /* Subscript to use for the next FGL. */
   int   new_fgl;                       /* If true, there is a new FGL.       */

   int   g1;                            /* Current subscript into the vector  */
   /*                                      of input ibd_graph gametes.        */
   int   m1;                            /* Subscript to current marker.       */

   /* Initialize the largest_fgl variable found in ibd_create.h to 0 before   */
   /* any of the get_index, add_hash_tail or print_table subroutines are ever */
   /* called just to ensure that largest_fgl cannot be set to some extremely  */
   /* large number just by chance.                                            */

   largest_fgl = 0;

   for (g1 = 0; g1 < num_gametes; ++g1)
   {
      curr_seglist = ibdgraph_gametes[g1].seglist;
      new_fgl = TRUE;

      for (m1 = 0; m1 < nsnps; ++m1)
      {
         /* Fast forward until we are at the right place along the input      */
         /* chromosome.  If we have a different FGL then we need to find the  */
         /* new FGL in the hashtable.                                         */

         while (curr_seglist->current->endpos < snp_pos[m1])
         {
            curr_seglist = curr_seglist->next;
            if (curr_seglist == NULL)
            {
               fprintf (stderr, "\n The input chromosome is too short!");
               fprintf (stderr, "\n Marker %d at position", m1);
               fprintf (stderr, " %d is not covered", snp_pos[m1]);
               fprintf (stderr, " by an FGL.\n\n");
               fflush (stderr);
            }
            new_fgl = TRUE;
         }

         if (new_fgl)
         {
            /* If we need to check the FGL, look it up in the hashtable and   */
            /* potentially update the value there, incrementing nextindex     */
            /* if this FGL is not already in the hashtable.                   */

            get_index (curr_seglist->current->fgl, hashtable, &nextindex);
            new_fgl = FALSE;
         }
      }
   }

   if (DebugLevel == 5)
      print_table (stderr, hashtable);

   /* Return the next FGL subscript to the calling routine.                   */

   *num_uniq_fgls = nextindex;

   return;
}

/******************************************************************************/
/*                                                                            */
/* Call the assign_alleles subroutine to assign alleles for each segment of   */
/* identity by descent (IBD) for each assigned founder genome label (FGL).    */
/*                                                                            */
/* The following are the original declarations of the arguments:              */
/*                                                                            */
/* int   nsnps;                            Number of SNPs in the snps matrix. */
/* int   nhaps;                            Number of haplotypes in the snps   */
/*                                         matrix.                            */
/*                                                                            */
/******************************************************************************/

char **assign_alleles (Hashnode ** hashtable, char **snps, int nsnps, int nhaps)
{
   Chromseglist *curr_seglist;          /* Current gamete segment list.       */

   char **result;                       /* The resulting alleles matrix.      */

   int   nextindex = 0;                 /* Subscript to use for the next FGL. */

   int   snp_array_col;                 /* The column subscript into the snps */
   /*                                      array.                             */
   int   fgl_dirty;                     /* If true, there is a new FGL.       */

   int   g1;                            /* Current subscript into the vector  */
   /*                                      of input ibd_graph gametes.        */
   int   h1;                            /* The current haplotype subscript.   */
   int   s1;                            /* The current SNP's subscript.       */

   fprintf (stdout, "\n Assigning haplotypes\n");
   fflush (stdout);

   /* Allocate the resulting alleles matrix.                                  */

   result = alloc_char2d (nsnps, num_gametes);

   /* Initialize the haplotype column subscript.                              */

   h1 = 0;

   /* Perform the following loop once for each gamete in the vector of input  */
   /* ibd_graph gametes.                                                      */

   for (g1 = 0; g1 < num_gametes; ++g1)
   {
      curr_seglist = ibdgraph_gametes[g1].seglist;
      fgl_dirty = TRUE;

      for (s1 = 0; s1 < nsnps; ++s1)
      {
         while (curr_seglist->current->endpos < snp_pos[s1])
         {
            /* Get to the right place along the chromosome.  If we move, we   */
            /* need to set fgl_dirty in order to look up the new FGL.         */

            curr_seglist = curr_seglist->next;
            if (curr_seglist == NULL)
            {
               fprintf (stderr, "\n");
               fprintf (stderr, " Terminal error: curr_seglist == NULL,");
               fprintf (stderr, "\n so an internal inconsistency in the");
               fprintf (stderr, " program has occurred.\n");
               give_up (ERR_qit);
            }
            fgl_dirty = TRUE;
         }

         if (fgl_dirty)
         {
            /* If we need to check the FGL, look it up in the hash table and  */
            /* possibly update the value there.                               */

            snp_array_col = get_index (curr_seglist->current->fgl, hashtable,
                                       &nextindex);
            fgl_dirty = FALSE;

            if (snp_array_col >= nhaps)
            {
               if (DebugLevel == 4)
                  print_table (stderr, hashtable);

               fprintf (stderr, "\n");
               fprintf (stderr, " Terminal error: snp_array_col >= nhaps,");
               fprintf (stderr, "\n so an internal inconsistency in the");
               fprintf (stderr, " program has occurred.\n");
               give_up (ERR_qit);
            }
         }

         /* Save the selected founder allele in the resulting alleles matrix. */

         result[s1][h1] = snps[s1][snp_array_col];
      }

      /* The current column of the resulting alleles matrix has been filled,  */
      /* so increment the subscript to the next column of the matrix.         */

      ++h1;
   }

   return result;
}

/******************************************************************************/
/*                                                                            */
/* Call the output_haplotypes subroutine to write the assigned haplotypes to  */
/* the fgl2haplo results file.                                                */
/*                                                                            */
/* The input 2 dimensional matrix of characters contains the alleles assigned */
/* for the haplotypes by the assign_alleles subroutine.  It consists of       */
/* nsnps (number of SNPs) rows and nhaps (number of haplotypes) columns.      */
/*                                                                            */
/******************************************************************************/

void output_haplotypes (FILE * oScorF, char **input, int nsnps, int nhaps)
{
   int   g1;                            /* Current subscript into the vector  */
   /*                                      of input ibd_graph gametes.        */

   int   s1;                            /* Current SNP's subscript.           */

   fprintf (stdout, "\n Outputting marker data as haplotypes\n\n");
   fflush (stdout);

   /* Perform the following block of code once for each gamete in the vector  */
   /* of input ibd_graph gametes.                                             */

   for (g1 = 0; g1 < num_gametes; ++g1)
   {
      if (OutHapsLabs)
      {
         fprintf (oScorF, "%s", ibdgraph_gametes[g1].name);
         fprintf (oScorF, " %d", ibdgraph_gametes[g1].which);
         fprintf (oScorF, " ");
      }

      for (s1 = 0; s1 < nsnps; ++s1)
      {
         /* Finally, print the allele to the output results file.             */

         if (OutWithSpac)
            fprintf (oScorF, " ");
         fprintf (oScorF, "%c", input[s1][g1]);
      }

      fprintf (oScorF, "\n");
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Call the output_genotypes subroutine to write the assigned haplotypes as   */
/* genotypic data to the fgl2haplo results file.                              */
/*                                                                            */
/* The input 2 dimensional matrix of characters contains the alleles assigned */
/* for the haplotypes by the assign_alleles subroutine.  It consists of       */
/* nsnps (number of SNPs) rows and nhaps (number of haplotypes) columns.      */
/*                                                                            */
/******************************************************************************/

void output_genotypes (FILE * oScorF, char **input, int nsnps, int nhaps)
{
   int   ninds;                         /* Number of individuals.             */

   int   i1;                            /* Current individual's subscript.    */
   int   s1;                            /* Current SNPS's subscript.          */
   int   h1;                            /* Current haplotype's subscript.     */
   int   h2;                            /* Next haplotype's subscript.        */

   fprintf (stdout, "\n Outputting marker data as genotypes\n\n");
   fflush (stdout);

   /* Start the output file with a Morgan "set markers ... data" statement.   */

   fprintf (oScorF, "set markers %d data\n", nsnps);
   ninds = min (num_gametes, nhaps) / 2;

   /* Perform the following block of code once for each individual.           */

   for (i1 = 0; i1 < ninds; ++i1)
   {
      h1 = i1 * 2;                      /* Compute subscripts to haplotypes.  */
      h2 = h1 + 1;

      if (OutHapsLabs)
      {
         if (strcmp (ibdgraph_gametes[h1].name, ibdgraph_gametes[h2].name))
         {
            fprintf (stderr, "\n");
            fprintf (stderr, " Terminal error: The names found for a pair of");
            fprintf (stderr, " gametes are not identical,\n");
            fprintf (stderr, " when h1 is %d,", h1);
            fprintf (stderr, " ibdgraph_gametes[h1].name is %s\n",
                     ibdgraph_gametes[h1].name);
            fprintf (stderr, "  and h2 is %d,", h2);
            fprintf (stderr, " ibdgraph_gametes[h2].name is %s\n",
                     ibdgraph_gametes[h1].name);

            fprintf (oScorF, "\n");
            give_up (ERR_qit);
         }
         else
            fprintf (oScorF, "%s", ibdgraph_gametes[h1].name);
      }

      for (s1 = 0; s1 < nsnps; ++s1)
      {
         /* Finally, print the pair of alleles to the output results file.    */

         if (OutWithSpac)
            fprintf (oScorF, "  ");
         fprintf (oScorF, "%c", input[s1][h1]);
         if (OutWithSpac)
            fprintf (oScorF, " ");
         fprintf (oScorF, "%c", input[s1][h2]);
      }

      fprintf (oScorF, "\n");
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The init_ss_ibd_data subroutine allocates and initializes all of the       */
/* vector of score_set_ibd_data structure elements that will be needed by the */
/* Morgan program that has called this subroutine.                            */
/*                                                                            */
/******************************************************************************/

void init_ss_ibd_data (FILE * oScorF)
{
   int  *index;                         /* Pointer to IBD state index.        */
   int  *number;                        /* Pointer to IBD state number.       */
   int  *label;                         /* Pointer to IBD state label.        */

   int   ss_size;                       /* Size of the current score set.     */
   int   ss_size_ndx;                   /* Current score set size subscript.  */

   int   maxlab;                        /* Number of labels.                  */
   int   numsts;                        /* Number of states.                  */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to current score set.    */
   int   s2;                            /* Subscript to current state.        */

   /* Allocate storage for the ss_ibd_data_structure vector.                  */

   ss_ibd_data = malloc (MxGam * sizeof (SS_IBD_DATA_STRUCTURE));
   alloc_chk (WHERE, ss_ibd_data);
   memset (ss_ibd_data, 0, MxGam * sizeof (SS_IBD_DATA_STRUCTURE));

   /* Check that for each score set entered by the user the corresponding     */
   /* element of the ss_ibd_data vector has been allocated and initialized.   */

   for (s1 = 0; s1 < NScSet[c1]; ++s1)  /* For each score set:                */
   {
      ss_size = NPrGam[c1][s1];         /* Get size of the current score set. */
      ss_size_ndx = ss_size - 1;        /* Set the score set size subscript.  */

      if (ss_ibd_data[ss_size_ndx].num_ss == 0)
      {
         ss_ibd_data[ss_size_ndx].ss_size = ss_size;

         maxlab = CDgg[ss_size];
         numsts = PTgg[ss_size];

         index = alloc_int1d (maxlab + 1);
         number = alloc_int1d (numsts);
         label = alloc_int1d (numsts + 1);

         /* Compute the IBD state vectors: index, number and label.           */

         index_ibd (ss_size, index, number, label);

         if (DebugLevel >= 6)
         {
            /* Output the index and label vectors to the oScorF file.         */

            fprintf (oScorF, "\n");

            for (s2 = 0; s2 < numsts; ++s2)
            {
               fprintf (oScorF, "%4d %4d %4d\n", s2, index[label[s2]],
                        label[s2]);
               if (s2 != index[label[s2]])
                  fprintf (oScorF,
                           " The line above this message looks wrong\n");
            }

            fprintf (oScorF, "\n");
            fflush (oScorF);
         }

         ss_ibd_data[ss_size_ndx].index = index;
         ss_ibd_data[ss_size_ndx].number = number;
         ss_ibd_data[ss_size_ndx].label = label;
      }

      ++ss_ibd_data[ss_size_ndx].num_ss;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_ss_ibd_data subroutine frees up the storage allocated by the      */
/* init_ss_ibd_data subroutine for use as the vector of score_set_ibd_data    */
/* structure elements.                                                        */
/*                                                                            */
/******************************************************************************/

void free_ss_ibd_data (void)
{
   int   ss_size_ndx;                   /* Current score set size subscript.  */

   for (ss_size_ndx = 0; ss_size_ndx < MxGam; ++ss_size_ndx)
   {
      free_1d ((void *) &ss_ibd_data[ss_size_ndx].index);
      free_1d ((void *) &ss_ibd_data[ss_size_ndx].number);
      free_1d ((void *) &ss_ibd_data[ss_size_ndx].label);
   }

   free_1d ((void *) &ss_ibd_data);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The fgl_marks subroutine determines the FGL's for a selected gamete,       */
/* passed as the first argument, at each of the SNP positions indicated by    */
/* the second and third arguments.  The results are stored in the vector      */
/* pointed to by the fourth argument.                                         */
/*                                                                            */
/* While doing this, the subroutine also finds how many characters would be   */
/* needed to print the the largest FGL and returns this number of characters  */
/* as its return value.                                                       */
/*                                                                            */
/******************************************************************************/

int fgl_marks (Chromseglist * gam, int nmarks, int *snppos, int *fgg)
{
   char  temp_buff[12];                 /* A temporary buffer to format the   */
   /*                                      current FGL to.                    */

   int   nf = 0;                        /* Character length of longest FGL.   */
   int   fgl;                           /* Copy of FGL at current marker.     */
   int   m1;                            /* Subscript to current marker.       */

   /* Assign the FGL's for the selected gamete.                               */

   for (m1 = 0; m1 < nmarks; ++m1)
   {
      /* Fast forward each FGL list through the SNP positions.                */
      while (gam->current->endpos < snppos[m1])
         gam = gam->next;

      /* Set the FGL assigned to the next marker.                             */
      if (gam)
      {
         fgl = gam->current->fgl;
         fgg[m1] = fgl;

         /* To make the data written to the output scores file line up        */
         /* nicely, find out how many characters are needed to write the      */
         /* largest FGL.                                                      */

         sprintf (temp_buff, "%d", fgl);
         nf = max (nf, strlen (temp_buff));
      }
   }

   return nf;
}

/******************************************************************************/
/*                                                                            */
/* The ibd_state subroutine converts the FGL at marker mk into an IBD state   */
/* vector consisting of ng gametes.  The resulting IBD state vector is        */
/* pointed to by the last argument.                                           */
/*                                                                            */
/******************************************************************************/

void ibd_state (int **fgl, int mk, int ng, int *states)
{
   int  *gg;

   int   nid;
   int   i;
   int   k;

   gg = alloc_int1d (ng);

   for (i = 0; i < ng; ++i)
   {
      gg[i] = fgl[i][mk];
      states[i] = 0;
   }

   states[0] = 1;
   nid = 1;

   for (i = 1; i < ng; ++i)
   {
      for (k = 0; k < i; ++k)
         if (gg[i] == gg[k])
         {
            states[i] = states[k];
            break;
         }

      if (k == i)
      {
         ++nid;
         states[i] = nid;
      }
   }

   free_1d ((void *) &gg);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The alloc_for_pairs subroutine allocates those variables needed to support */
/* the "set proband gametes all individual pairs" parameter statement.  This  */
/* option requires only one score set, so the s1 variable will always be set  */
/* to 0.                                                                      */
/*                                                                            */
/* For right now this subroutine is handling only a single component, so the  */
/* c1 variable will always be set to 0.                                       */
/*                                                                            */
/******************************************************************************/

void alloc_for_pairs (void)
{
   int   c1;                            /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to score set (also 0).   */
   int   g1;                            /* Subscript to proband gamete.       */

   /* To support the "all individual pairs" option we will need storage for   */
   /* only one score set.                                                     */

   NScCmp = 1;
   NScCmpAlloc = NScCmp;

   NScSet = malloc (NScCmp * sizeof (int));
   alloc_chk (WHERE, NScSet);
   ScSetNo = malloc (NScCmp * sizeof (int *));
   alloc_chk (WHERE, ScSetNo);
   NPrGam = malloc (NScCmp * sizeof (int *));
   alloc_chk (WHERE, NPrGam);
   ProbNam = malloc (NScCmp * sizeof (char ***));
   alloc_chk (WHERE, ProbNam);
   ProbIx = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, ProbIx);
   ProbGam = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, ProbGam);

   for (c1 = 0; c1 < NScCmp; ++c1)
   {
      NScSet[c1] = 1;

      ScSetNo[c1] = malloc (NScSet[c1] * sizeof (int));
      alloc_chk (WHERE, ScSetNo[c1]);
      NPrGam[c1] = malloc (NScSet[c1] * sizeof (int));
      alloc_chk (WHERE, NPrGam[c1]);
      ProbNam[c1] = malloc (NScSet[c1] * sizeof (char **));
      alloc_chk (WHERE, ProbNam);
      ProbIx[c1] = malloc (NScSet[c1] * sizeof (int *));
      alloc_chk (WHERE, ProbIx[c1]);
      ProbGam[c1] = malloc (NScSet[c1] * sizeof (int *));
      alloc_chk (WHERE, ProbGam[c1]);
   }

   for (c1 = 0; c1 < NScCmp; ++c1)
   {
      for (s1 = 0; s1 < NScSet[c1]; ++s1)
      {
         ScSetNo[c1][s1] = 1;           /* Start score set number at 1.       */

         /* The score set always consists of 4 gametes.                       */
         MxGam = NPrGam[c1][s1] = 4;

         ProbNam[c1][s1] = malloc (NPrGam[c1][s1] * sizeof (char *));
         alloc_chk (WHERE, ProbNam[c1][s1]);
         ProbIx[c1][s1] = malloc (NPrGam[c1][s1] * sizeof (int));
         alloc_chk (WHERE, ProbIx[c1][s1]);
         ProbGam[c1][s1] = malloc (NPrGam[c1][s1] * sizeof (int));
         alloc_chk (WHERE, ProbGam[c1][s1]);

         for (g1 = 0; g1 < NPrGam[c1][s1]; ++g1) /* For each proband gamete.  */
         {
            ProbNam[c1][s1][g1] = malloc ((MAXNAMLEN + 1) * sizeof (char));
            alloc_chk (WHERE, ProbNam[c1][s1][g1]);

            ProbGam[c1][s1][g1] = g1 % 2;
         }
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The init_ibd_lengths subroutine initializes the four integer length        */
/* variables which are declared in the ibd_haplo_directory.h header file and  */
/* are used by both the compute_ibd_probs and the gen_fgl2ibd_labels          */
/* subroutines.                                                               */
/*                                                                            */
/******************************************************************************/

void init_ibd_lengths (void)
{
   char  temp_buff[12];                 /* A temporary buffer to format the   */
   /*                                      highest selected marker number and */
   /*                                      highest selected score set number  */
   /*                                      to.                                */

   int   highest_ss_number;             /* Computed highest score set number. */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to current score set.    */
   int   i1;                            /* Subscript to current individual.   */
   int   g1;                            /* Subscript to current gamete.       */

   int   l_1;                           /* Provisional name length.           */

   /* Initialize the integer variable, nc, to the number of characters in the */
   /* longest name of an observed individual in the dataset.                  */

   nc = 0;
   if (NObs > 0)
   {
      for (i1 = 0; i1 < NObs; ++i1)
      {
         l_1 = strlen (ObsNam[i1]);
         nc = l_1 > nc ? l_1 : nc;
      }
   }

   /* If there are no observed individuals, then initialize nc to the number  */
   /* of characters in the longest name found in any input ibd_graph gamete.  */

   else if (ibdgraph_gametes != NULL)
   {
      for (g1 = 0; g1 < num_gametes; ++g1)
      {
         l_1 = strlen (ibdgraph_gametes[g1].name);
         nc = l_1 > nc ? l_1 : nc;
      }
   }

   /* To make the data written to the output scores file line up nicely, find */
   /* out how many characters are needed to write the marker number of the    */
   /* highest numbered selected marker.                                       */

   sprintf (temp_buff, "%d", MarSel[NMarks - 1]);
   nm = strlen (temp_buff);

   /* We will also need to know how many characters are needed to write the   */
   /* highest numbered proband gamete score set number to.                    */

   if (!DoingAllPairs)
   {
      nsn = 0;
      for (s1 = 0; s1 < NScSet[c1]; ++s1)
      {
         sprintf (temp_buff, "%d", ScSetNo[c1][s1]);
         nsn = max (nsn, strlen (temp_buff));
      }
   }
   else
   {
      highest_ss_number = max ((num_names * (num_names - 1) / 2), 1);
      sprintf (temp_buff, "%d", highest_ss_number);
      nsn = strlen (temp_buff);
   }

   /* Compute how many characters are needed to write the largest number of   */
   /* possible IBD states for any score set.  This will be for any score set  */
   /* with the most proband gametes.  For aesthetic reasons we don't want     */
   /* this to be less than two characters.                                    */

   sprintf (temp_buff, "%d", PTgg[MxGam]);
   nst = max (2, strlen (temp_buff));

   return;
}

/******************************************************************************/
/*                                                                            */
/* The make_names_vector subroutine is called to initialize the indiv_names   */
/* vector to contain a sorted vector of the unique names found in the input   */
/* ibdgraph gametes data file.                                                */
/*                                                                            */
/* This will require two passes through the ibdgraph_gametes vector.  In the  */
/* first pass the number of unique names is counted.  This is needed so that  */
/* the storage needed for the names can be allocated.  Then in a second pass  */
/* the unique names found in the ibdgraph_gametes vector can be copied.       */
/*                                                                            */
/******************************************************************************/

void make_names_vector (void)
{
   int   copy;                          /* Set TRUE whenever the current name */
   /*                                      needs to be copied to indiv_names. */

   int   name_ndx;                      /* Subscript to the next name.        */
   int   gam_ndx;                       /* Subscript to the current gamete.   */

   int   g0;                            /* Subscript to the previous gamete.  */
   int   g1;                            /* Subscript to the current gamete.   */

   /* Count the number of unique names found in the ibdgraph_gametes vector.  */

   num_names = 1;                       /* Always count the first name.       */

   for (g1 = 1; g1 < num_gametes; ++g1)
   {
      g0 = g1 - 1;
      if (strcmp (ibdgraph_gametes[g0].name, ibdgraph_gametes[g1].name) != 0)
         ++num_names;
   }

   /* Allocate the storage needed for a vector of all the unique names found  */
   /* in the input ibd_graph data.                                            */

   indiv_names = alloc_char2d (num_names, MAXNAMLEN + 1);

   copy = TRUE;
   name_ndx = 0;
   gam_ndx = 0;

   while (gam_ndx < num_gametes)
   {
      /* The copy variable will always be set to FALSE except for the first   */
      /* time through this loop.  We always want the first name to be copied, */
      /* so there is no need to make the call to strcmp for the first name.   */

      if (copy == FALSE)
      {
         g0 = gam_ndx - 1;
         g1 = gam_ndx;
         copy = strcmp (ibdgraph_gametes[g0].name, ibdgraph_gametes[g1].name);
      }

      if (copy)
      {
         /* Copy the name from the ibdgraph_gametes vector to the vector of   */
         /* unique names that we are creating.                                */

         strcpy (indiv_names[name_ndx], ibdgraph_gametes[gam_ndx].name);

         copy = FALSE;
         ++name_ndx;
         ++gam_ndx;
      }
      else
         ++gam_ndx;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The set_for_pair subroutine is called to set those variables which need to */
/* be dynamically set during the execution of an IBD_Haplo program while      */
/* supporting the "set proband gametes all individual pairs" parameter        */
/* statement.  This option requires only one score set, so the s1 variable    */
/* will always be set to 0.                                                   */
/*                                                                            */
/* Since distinct components are not supported by IBD_Haplo programs, the     */
/* c1 variable will always be set to 0.                                       */
/*                                                                            */
/* The two arguments are declared in the calling routine as:                  */
/*                                                                            */
/* int   i1;                             * Subscript of the first individual  */
/*                                         in a pair of individuals.          */
/*                                                                            */
/* int   i2;                             * Subscript of the second individual */
/*                                         in a pair of individuals.          */
/*                                                                            */
/* Since this subroutine is only called when running the all pairs case, so:  */
/* for this subroutine we know that NScCmp has been set to 1,                 */
/*                             that NScSet[c1] has been set to 1 and          */
/*                             that NPrGam[c1][s1] has been set to 4.         */
/*                                                                            */
/******************************************************************************/

void set_for_pair (int i1, int i2)
{
   int   c1;                            /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to score set (also 0).   */
   int   g1;                            /* Subscript to current gamete.       */

   for (c1 = 0; c1 < NScCmp; ++c1)
   {
      for (s1 = 0; s1 < NScSet[c1]; ++s1)
      {
         for (g1 = 0; g1 < NPrGam[c1][s1]; ++g1)
         {
            if (g1 <= 1)
            {
               strncpy (ProbNam[c1][s1][g1], indiv_names[i1], MAXNAMLEN);
               ProbIx[c1][s1][g1] = i1;
            }
            else
            {
               strncpy (ProbNam[c1][s1][g1], indiv_names[i2], MAXNAMLEN);
               ProbIx[c1][s1][g1] = i2;
            }
         }
      }
   }

   return;
}
