/******************************************************************************/
/*                                                                            */
/* This file contains a collection of subroutines used by the ibd_create      */
/* subprograms.                                                               */
/*                                                                            */
/******************************************************************************/

#include "incs.h"
#include "ibd_create.h"

#define HASHPRIME 337
#define MYHASH(x) ((x) % HASHPRIME)

/******************************************************************************/
/*                                                                            */
/* The add_hash_tail subroutine adds a new node to the linked list of         */
/* hashnodes passed as the first argument to this subroutine.                 */
/*                                                                            */
/* As a side effect of this subroutine, once the hashtable has been filled,   */
/* the largest FGL found in the data will be saved in largest_fgl.            */
/*                                                                            */
/******************************************************************************/

Hashnode **add_hash_tail (Hashnode ** tailaddr, int newfgl, int newindex)
{
   Hashnode *out;

   out = malloc (sizeof (Hashnode));
   alloc_chk (WHERE, out);

   out->fgl = newfgl;
   out->index = newindex;
   out->next = NULL;

   largest_fgl = max (largest_fgl, newfgl);

   *tailaddr = out;

   return &((*tailaddr)->next);
}

/******************************************************************************/
/*                                                                            */
/* The alloc_hashtable subroutine allocates and zeroes the storage needed for */
/* a hashtable to be used for referring to the FGLs.                          */
/*                                                                            */
/* The hashtable is just an array of HASHPRIME pointers to hashnodes, indexed */
/* by the hashes.  FGLs which have the same hash are stored as a linked list. */
/*                                                                            */
/******************************************************************************/

Hashnode **alloc_hashtable ()
{
   Hashnode **out;                      /* Pointer to the output hashtable.   */

   int   ht_size;                       /* Total number of bytes needed for   */
   /*                                      the hashtable.                     */

   ht_size = HASHPRIME * sizeof (Hashnode);

   out = malloc (ht_size);
   alloc_chk (WHERE, out);

   memset (out, 0, ht_size);

   return out;
}

int get_index (int fgl, Hashnode ** hashtable, int *nextindex)
/* Looks up an FGL in the hash table and adds a new
 * node if it's not found; nextindex is the next available
 * input gamete, so it gets assigned and incremented
 * if we add a new node. */
{
   Hashnode **curr;

   int   b = MYHASH (fgl);

   curr = &hashtable[b];
   while (*curr)
   {
      if ((*curr)->fgl == fgl)
      {
         return (*curr)->index;
      }
      curr = &((*curr)->next);
   }

   /* If we get here, we have not found the FGL.                              */
   add_hash_tail (curr, fgl, (*nextindex)++);

   return (*curr)->index;
}

/******************************************************************************/
/*                                                                            */
/* The print_table subroutine prints out the contents of the FGL hashtable.   */
/*                                                                            */
/* The output consists of the current subscript into the hashtable, the       */
/* current FGL, a count of number of unique FGL's and the current assigned    */
/* index.                                                                     */
/*                                                                            */
/* The largest_fgl variable needs to be set before calling print_table.       */
/*                                                                            */
/******************************************************************************/

void print_table (FILE * outfile, Hashnode ** hashtable)
{
   Hashnode *curr_fgl;

   char  temp_buff[12];                 /* A temporary buffer to format the   */
   /*                                      largest FGL to.                    */

   int   new_hash_code;                 /* Set to TRUE when a new hash_code   */
   /*                                      is being worked on.                */

   int   uniq_fgls_cnt;                 /* Count of number of unique FGL's.   */
   int   l_largest_fgl;                 /* Character length of largest FGL.   */
   int   x1;

   fprintf (outfile, "\n");             /* Skip a line before debugging info. */

   /* Determine how many characters will be needed to print the largest FGL   */
   /* found in the data.                                                      */

   sprintf (temp_buff, "%d", largest_fgl);
   l_largest_fgl = strlen (temp_buff);

   uniq_fgls_cnt = 0;

   for (x1 = 0; x1 < HASHPRIME; ++x1)
   {
      new_hash_code = TRUE;

      curr_fgl = hashtable[x1];
      while (curr_fgl)
      {
         ++uniq_fgls_cnt;

         if (new_hash_code)
         {
            fprintf (outfile, "%4d", x1);
            new_hash_code = FALSE;
         }
         else
            fprintf (outfile, "    ");

         fprintf (outfile, "  %*d", l_largest_fgl, curr_fgl->fgl);
         fprintf (outfile, "  %*d", l_largest_fgl, uniq_fgls_cnt);

         fprintf (outfile, "  %*d", l_largest_fgl, curr_fgl->index);
         fprintf (outfile, "\n");

         curr_fgl = curr_fgl->next;
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The init_snp_pos subroutine is a short subroutine which can be called to   */
/* initialize the snp_pos vector.                                             */
/*                                                                            */
/******************************************************************************/

void init_snp_pos (int nsnps)
{
   double intermediate_snp_pos;         /* Current SNP's position before      */
   /*                                      rounding to the nearest integer.   */

   int   m1;                            /* Subscript to current marker.       */

   /* Allocate storage for the SNP positions vector.                          */

   snp_pos = alloc_int1d (nsnps);

   /* The fgl_marks subroutine expects the snp_pos vector to contain physical */
   /* map distances (in number of base pairs), so here we multiply the Morgan */
   /* map positions (which were input as centimorgans) by 1 million each.     */

   for (m1 = 0; m1 < nsnps; ++m1)
   {
      /* In MarkPos it doesn't matter whether I refer to the female or male   */
      /* marker position, so the first subscript can be either 0 or 1.        */
      intermediate_snp_pos = FirstMarkPos + MarkPos[0][m1];

      /* Perform an arithmetic rounding of the result to the nearest integer  */
      /* in order to prevent any truncation that would otherwise occur.       */
      snp_pos[m1] = floor ((intermediate_snp_pos * 1E6) + 0.5) / CMperMbp;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_hashtable subroutine frees up the storage allocated by the        */
/* fgl2haplo subprogram for the hash table.                                   */
/*                                                                            */
/******************************************************************************/

void free_hashtable (Hashnode ** hashtable)
{
   Hashnode *curr_hashnode;             /* Points to current hashtable entry. */
   Hashnode *next_hashnode;             /* Pointer to next hashtable entry.   */

   int   h1;                            /* Current subscript into hashtable.  */

   for (h1 = 0; h1 < HASHPRIME; ++h1)
   {
      curr_hashnode = hashtable[h1];

      while (curr_hashnode != NULL)
      {
         next_hashnode = curr_hashnode->next;
         free (curr_hashnode);
         curr_hashnode = next_hashnode;
      }
   }

   free (hashtable);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Read in all the input ibd_graph gamete data created by the simpop_fgl      */
/* subprogram.  This will require two passes through the input file.  In the  */
/* first pass the number of gametes is counted.  This is needed so that an    */
/* output array of gametes can be allocated.  Then in a second pass the input */
/* ibd_graph gamete data needs to be read into the array.                     */
/*                                                                            */
/* The p_rec and line_num variables are not really needed currently.  When    */
/* this subroutine is modified to perform better error checking of the input  */
/* data, these variables should prove to be useful.       SML  Sep 2, 2014    */
/*                                                                            */
/******************************************************************************/

void read_gamete_data (FILE * fp, int *num_gams, Gamete ** result)
{
   Gamete *out;

   char  nxt_lin[MAXLINLEN + 2];
   char  gamname[MAXNAMLEN + 1];
   char *p_rec;

   /* The p_rec and line_num variables should be used during the second time  */
   /* through the input file where the input should be read into a string, so */
   /* that error checks can be made regarding the input.                      */

   long  line_num;                      /* Line number of line just read from */
   /*                                      the input ibd_graph gamete file.   */

   int   numgams;                       /* A local variable for keeping the   */
   /*                                      total number of gametes found in   */
   /*                                      the input ibd_graph gamete file.   */

   int   l_name;                        /* Length of the current gamname.     */

   int   g1;                            /* Subscript to the current gamete.   */

   /* Count the number of lines of input ibd_graph gamete data that we will   */
   /* need to read in.                                                        */

   numgams = 0;
   line_num = 0;

   while ((p_rec = fgets (nxt_lin, MAXLINLEN + 1, fp)) != NULL)
   {
      ++numgams;

      ++line_num;                       /* Increment the line number.         */
      if (DebugLevel == 6)
         printf (" Input gamete file line number %ld: %s\n", line_num, p_rec);
   }

   /* Allocate enough memory for the input ibd_graph gamete data.             */

   out = (Gamete *) malloc (numgams * sizeof (Gamete));
   alloc_chk (WHERE, out);

   /* Now reread the input ibd_graph gamete data into the output vector.      */

   rewind (fp);

   for (g1 = 0; g1 < numgams; ++g1)
   {
      /* Read in the input ibd_graph gamete data for the current gamete.      */

      /* Copy the identifying information: subscript, individual name and     */
      /* which gamete (maternal or paternal) into the output array.           */

      out[g1].index = g1;               /* Save the subscript.                */

      fscanf (fp, "%s", gamname);       /* Name of the individual to whom     */
      /*                                   the gamete belongs.                */
      l_name = strlen (gamname) + 1;    /* Leave room for the null character. */
      out[g1].name = alloc_char1d (l_name);
      strcpy (out[g1].name, gamname);

      fscanf (fp, "%*d %*d %*d %*d");   /* Skip the data which is not needed. */

      fscanf (fp, "%d", &out[g1].which); /* Set which gamete.                 */

      /* Now store the input ibd_graph gamete data into the output array.     */

      out[g1].seglist = read_chrom (fp);
   }

   *num_gams = numgams;
   *result = out;

   return;
}

Chromseglist *read_chrom (FILE * fp)
/* read in junction data starting AFTER the id #s at start of the line */
{
   Chromseglist *out = NULL;
   Chromseglist **tailaddr;
   int   tempfgl;
   long long tempendpos;

   tailaddr = &out;

   /* kludge: fscanf zooms past newline characters, so we step back one
    * character after every read and check to see if the last
    * skipped whitespace was a newline. in theory the line can be arbitrarily
    * large so we can't buffer an entire line. the "[]" has to be
    * in the fscanf pattern. fscanf contains deep mysteries. */
   do
   {
      fscanf (fp, "%d %lld []", &tempfgl, &tempendpos);
      tailaddr = add_chrom_tail (tailaddr, tempfgl, tempendpos);
      fseek (fp, -1, SEEK_CUR);
   }
   while (getc (fp) != '\n');
   return out;
}

/******************************************************************************/
/*                                                                            */
/* The free_gamete_data subroutine is called to free up any storage allocated */
/* for the purpose of storing the data read in from the input ibd_graph       */
/* gamete data file produced as the output of the simpop_fgl subprogram.      */
/*                                                                            */
/******************************************************************************/

void free_gamete_data (int num_gams, Gamete * gametes)
{
   Chromseglist *curr_seglist;
   Chromseglist *next_seglist;

   int   g1;                            /* Subscript into the input ibd_graph */
   /*                                      gamete data vector.                */

   for (g1 = 0; g1 < num_gams; ++g1)
   {
      free (gametes[g1].name);

      curr_seglist = gametes[g1].seglist;

      while (curr_seglist != NULL)
      {
         next_seglist = curr_seglist->next;
         free (curr_seglist->current);
         free (curr_seglist);
         curr_seglist = next_seglist;
      }
   }

   free (gametes);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The input_next_alphanum subroutine continues reading characters from the   */
/* input file until the next alphanumeric character is read in.  The return   */
/* code from the last call to the fscanf routine is also returned by this     */
/* subroutine to the caller.                                                  */
/*                                                                            */
/* So long as the return code from fscanf is a 1, this subroutine makes sure  */
/* that whatever character it returns as next_char is an alphanumeric         */
/* character.  If the return code from fscanf is not a 1, then next_char is   */
/* not modified.  So next_char cannot be a control character, such as '\n'.   */
/*                                                                            */
/******************************************************************************/

int input_next_alphanum (FILE * inp_file, char *next_char)
{
   char  buff;                          /* A single character used for input. */

   int   rc_fscanf;                     /* The return code from fscanf.       */
   int   rc_isalnum;                    /* The return code from isalnum.      */

   do                                   /* Perform a "do forever" loop.       */
   {
      /* Get the next non-blank character from the designated input file.     */

      rc_fscanf = fscanf (inp_file, "%1c", &buff);

      /* If the return code from fscanf is not 1 then there was no character  */
      /* returned from fscanf since we only asked for one character of input. */
      /* So there is no need to check if it is an alphanumeric character.     */

      if (rc_fscanf != 1)
         break;

      /* Otherwise, check to make sure that the character returned by fscanf  */
      /* is an alphanumeric character.  If it is, copy the character to       */
      /* next_char and break from the "do forever" loop.                      */

      rc_isalnum = isalnum (buff);
      if (rc_isalnum)
      {
         memcpy (next_char, &buff, 1);
         break;
      }
   }
   while (TRUE);

   return rc_fscanf;
}

/******************************************************************************/
/*                                                                            */
/* The populate_snps subroutine is called to read data from the input file    */
/* passed as the second argument into a "snps" result array, which is both    */
/* allocated and set by this subroutine.                                      */
/*                                                                            */
/* This subroutine reads in one character at a time from the input file.      */
/*                                                                            */
/* The result array will always be allocated in memory so that each row       */
/* represents a SNP and each column represents a haplotype.                   */
/*                                                                            */
/* The first argument to this subroutine is a null-terminated string.  This   */
/* string consists of option code letters and is used to indicate which parts */
/* of the subroutine the caller wants executed.                               */
/*                                                                            */
/* The option code letters, 'S' or 'H', reflect the order the data is read in */
/* from the input file, the letter reflecting which subscript is being        */
/* incremented most rapidly (ie, the rightmost subscript).  One or the other  */
/* of these option code letters is required by the subroutine.  If neither is */
/* present, the subroutine does not populate the results array and returns to */
/* the caller with an error return code of 1001.                              */
/*                                                                            */
/* If the 'E' option code letter is passed, then this subroutine is asked to  */
/* make sure that the data from the input file ends when it would be expected */
/* to end.  If the number of SNPs and number of haplotypes specified in the   */
/* metadata file are correct for the input file, the data from the input file */
/* should be exhausted precisely when the main input loops reach their end,   */
/* so if this subroutine is able to read another character from the input     */
/* file it is an error and the subroutine exits with a return code of 1002.   */
/*                                                                            */
/******************************************************************************/

int populate_snps (const char *options, FILE * inp_file, const int nsnps,
                   const int nhaps, char ***result)
{
   char  buff = ' ';                    /* A single character used for input. */
   char **out;                          /* Local copy of the result array.    */

   int   s1;                            /* Current SNP subscript.             */
   int   h1;                            /* Current haplotype subscript.       */

   int   rc;                            /* Return code we received.           */

   out = alloc_char2d (nsnps, nhaps);

   if (strchr (options, 'S'))
   {
      for (h1 = 0; h1 < nhaps; ++h1)
      {
         for (s1 = 0; s1 < nsnps; ++s1)
         {
            rc = input_next_alphanum (inp_file, &buff);
            if (rc != 1)
            {
               *result = out;
               return rc;
            }
            out[s1][h1] = buff;
         }
      }
   }

   else if (strchr (options, 'H'))
   {
      for (s1 = 0; s1 < nsnps; ++s1)
      {
         for (h1 = 0; h1 < nhaps; ++h1)
         {
            rc = input_next_alphanum (inp_file, &buff);
            if (rc != 1)
            {
               *result = out;
               return rc;
            }
            out[s1][h1] = buff;
         }
      }
   }

   else
   {
      /* Neither a 'S' option code nor a 'H' option code were passed to the   */
      /* subroutine, so return to the caller with an error return code.       */

      free_2d ((void *) &out);
      *result = NULL;

      return 1001;
   }

   *result = out;

   if (strchr (options, 'E'))
   {
      rc = input_next_alphanum (inp_file, &buff);
      if (rc == 1)
      {
         /* Then there is more data in the input file, so return to the       */
         /* caller with an error return code.                                 */

         return 1002;
      }
   }

   /* Otherwise, the populating of the result array has been successful.      */

   return 0;
}

Chromseglist *get_ibd (Chromseglist * first, Chromseglist * second)
/* the algorithm here is roughly:
 * - set a pointer to the first node of each chromosome
 * - set last_end == 0
 * 1. check for equality of fgl
 * 1. a. if yes, we have entered an IBD segment
 *       keep advancing both pointers
 *      until they are different (meaning we
 *      have left the IBD segment)
 *    b. the IBD seg starts at "last_end" and ends
 *      at the lesser of the two pointers
 * 2. set "last_end" to the endpos of the most recently
 *    moved pointer (pointed to by curr_ptr)
 * 3. advance the lesser of the two pointers
 * 4. go to 1
 *
 * The idea is to zig-zag up the pair of chromosomes
 * from this:
 * ______*______________________
 *      /
 * ____/________________________
 *    *
 * to this:
 * ______*______________________
 *        \
 * ________\____________________
 *          *
 *
 * comparing fgl at each step
 *
 * Seems to work in basic cases but need to make sure
 * it's working at beginning or end of chromosome.
 */
{
   Chromseglist *out = NULL;
   Chromseglist **tail_ibd;
   Chromseglist *curr_first, *curr_second;
   Chromseglist **curr_ptr;
   Chromseglist *prev_ptr = NULL;
   long long last_end;

   tail_ibd = &out;
   curr_first = first;
   curr_second = second;
   curr_ptr = &curr_first;
   last_end = 0;

   while (*curr_ptr)
   {
      /* select smaller junction */
      if (curr_first->current->endpos < curr_second->current->endpos)
         curr_ptr = &curr_first;
      else
         curr_ptr = &curr_second;

      /* have we started an IBD segment? */
      if (curr_first->current->fgl == curr_second->current->fgl)
      {
         while (curr_first->next && curr_second->next &&
                curr_first->current->endpos == curr_second->current->endpos)
            /* advance through the IBD segment */
         {
            curr_first = curr_first->next;
            curr_second = curr_second->next;
         }

         /* IBD segment ends at smaller of first two unequal breaks */
         if (curr_first->current->endpos < curr_second->current->endpos)
            curr_ptr = &curr_first;
         else
            curr_ptr = &curr_second;

         tail_ibd = add_chrom_tail (tail_ibd, 0, last_end);
         tail_ibd = add_chrom_tail (tail_ibd, 1, (*curr_ptr)->current->endpos);
      }

      /* advance the smaller junction */
      last_end = (*curr_ptr)->current->endpos;
      prev_ptr = (*curr_ptr);
      (*curr_ptr) = (*curr_ptr)->next;
   }
   /* this is sloppy: tack on a no-ibd segment at the end,
    * with endpoint the end of the chromosome.
    * if there's already a yes-ibd segment there,
    * no matter, we'll never get there */
   tail_ibd = add_chrom_tail (tail_ibd, 0, prev_ptr->current->endpos);
   return out;
}

Chromseglist **add_chrom_tail (Chromseglist ** tailaddr, int newfgl,
                               long long newendpos)
{
   /* Takes a Chromseglist pointer by reference, creates a new node,
    * and changes the passed pointer to point to the new node. */

   Chromseglist *out;

   out = (Chromseglist *) malloc (sizeof (Chromseglist));
   out->current = (Chromseg *) malloc (sizeof (Chromseg));
   out->current->fgl = newfgl;
   out->current->endpos = newendpos;
   out->next = NULL;
   *tailaddr = out;

   return &((*tailaddr)->next);
}

int  *ibd_fifteen (Gamete mat1, Gamete pat1, Gamete mat2, Gamete pat2,
                   int nmarks, int *snp_pos)
{
   /* We can identify the haplotype-based IBD status (15 possible states)
    * by testing each pair of gametes.  Each pattern of binary
    * IBD states maps to an integer in [0,2^6-1].  We then map
    * that integer to the ibd state 1 through 15, inclusive.
    *
    * state pattern  integer
    *  1    1 1 1 1  63
    *  2    1 1 2 2  33
    *  3    1 1 1 2  11
    *  4    1 1 2 1  21
    *  5    1 1 2 3  1
    *  6    1 2 1 1  38
    *  7    1 2 2 2  56
    *  8    1 2 3 3  32
    *  9    1 2 1 2  18
    * 10    1 2 2 1  12
    * 11    1 2 1 3  2
    * 12    1 2 3 1  4
    * 13    1 2 2 3  8
    * 14    1 2 3 2  16
    * 15    1 2 3 4  0
    *
    */

   Chromseglist *tests[6];
   Chromseglist *curr[6];

   int  *out;
   int   code;
   int   i;
   int   j;

   out = alloc_int1d (nmarks);

   /* Assemble Chromseglists with FGLs either 1 or 0, indicating pairwise IBD */
   /* segments.                                                               */

   tests[0] = get_ibd (mat1.seglist, pat1.seglist);
   tests[1] = get_ibd (mat1.seglist, mat2.seglist);
   tests[2] = get_ibd (mat1.seglist, pat2.seglist);
   tests[3] = get_ibd (pat1.seglist, mat2.seglist);
   tests[4] = get_ibd (pat1.seglist, pat2.seglist);
   tests[5] = get_ibd (mat2.seglist, pat2.seglist);

   for (j = 0; j < 6; ++j)
      curr[j] = tests[j];

   for (i = 0; i < nmarks; ++i)
   {
      code = 0;
      for (j = 0; j < 6; ++j)
      {
         /* fast forward each IBD list */
         while (curr[j]->current->endpos < snp_pos[i])
         {
            curr[j] = curr[j]->next;
         }

         /* add test status times 2^j */
         code += curr[j]->current->fgl << j;
      }
      switch (code)
      {
          case 63:
             out[i] = 1;
             break;
          case 33:
             out[i] = 2;
             break;
          case 11:
             out[i] = 3;
             break;
          case 21:
             out[i] = 4;
             break;
          case 1:
             out[i] = 5;
             break;
          case 38:
             out[i] = 6;
             break;
          case 56:
             out[i] = 7;
             break;
          case 32:
             out[i] = 8;
             break;
          case 18:
             out[i] = 9;
             break;
          case 12:
             out[i] = 10;
             break;
          case 2:
             out[i] = 11;
             break;
          case 4:
             out[i] = 12;
             break;
          case 8:
             out[i] = 13;
             break;
          case 16:
             out[i] = 14;
             break;
          case 0:
             out[i] = 15;
             break;
          default:
             fprintf (stderr, "Unrecognized IBD pattern");
      }
   }

   for (j = 0; j < 6; ++j)
      free_chromseglist (tests[j]);

   return out;
}

int  *ibd_nine (int *ibd_fifteen, int nmarks)
{
   /* The genotype-based statuses are functions of the 15-state statuses.     */

   int  *out;
   int   i;

   out = alloc_int1d (nmarks);

   for (i = 0; i < nmarks; ++i)
   {
      switch (ibd_fifteen[i])
      {
          case 1:
             out[i] = 1;
             break;
          case 2:
             out[i] = 2;
             break;
          case 3:
          case 4:
             out[i] = 3;
             break;
          case 5:
             out[i] = 4;
             break;
          case 6:
          case 7:
             out[i] = 5;
             break;
          case 8:
             out[i] = 6;
             break;
          case 9:
          case 10:
             out[i] = 7;
             break;
          case 11:
          case 12:
          case 13:
          case 14:
             out[i] = 8;
             break;
          case 15:
             out[i] = 9;
             break;
      }
   }

   return out;
}

/******************************************************************************/
/*                                                                            */
/* The comparGams subroutine was written for use as a comparison subroutine   */
/* by the standard C library qsort and bsearch subroutines.  This subroutine  */
/* is called to specifically compare gamete array structure elements.  These  */
/* are declared in the ibd_create.h header file.                              */
/*                                                                            */
/******************************************************************************/

int comparGams (const void *key, const void *elem)
{
   P_Gamete p_array_element;
   P_Gamete p_key;

   int   result;

   p_array_element = (P_Gamete const) elem;
   p_key = (P_Gamete const) key;

   /* First compare the name components in the two Gamete structures.         */

   result = strncmp (p_key->name, p_array_element->name, MAXNAMLEN);

   /* If result is 0 then the names are identical, so we also need to         */
   /* compare the types of gamete (0 -> maternal, 1 -> paternal) components   */
   /* in the two Gamete structures.                                           */

   if (result == 0)
      result = p_key->which - p_array_element->which;

   return result;
}
