/******************************************************************************/
/*                                                                            */
/* Created from simu.c and new_simu.c                                         */
/*                                                                            */
/******************************************************************************/

#include "parsie.h"
#include "rans.h"
#include "ibd_create.h"

/* TO DO:
 * - malloc checks.  Calls to alloc_chk need to be added.                     */

static FILE *oScorF = NULL;

static int labelcounter;

Breaklist *makebreaks (long long chrlength, double rate)
/* rate is still a double, expressed as breaks per chrlength */
{
   Breaklist *current, *head = NULL;

   long long exprand, runsum = 0;

   while (TRUE)
      /* constructed backwards by keeping track of head */
   {
      /* exponential RV */
      /* this would get wacky if chrlength were unsigned */
      exprand = (long long) floor (-chrlength * log (rans ()) / rate);
      runsum += exprand;
      if (runsum > chrlength)
         break;
      current = (Breaklist *) malloc (sizeof (Breaklist));
      current->addr = chrlength - runsum;
      current->next = head;
      head = current;
   }

   return head;
}

Ibdlist **add_ibd_tail (Ibdlist ** tailaddr, long long newstart,
                        long long newend)
    /* takes an Ibdlist pointer by reference, creates a new node,
     * and changes the passed pointer to point to the new node */
{
   Ibdlist *out;

   out = (Ibdlist *) malloc (sizeof (Ibdlist));
   out->start = newstart;
   out->end = newend;
   out->next = NULL;
   *tailaddr = out;

   return &((*tailaddr)->next);
}

Chromseglist *recombine (Person x, Breaklist * newbreaks)
    /* creates a new Chromseglist pointer from a Person's maternal
     * and paternal chromosomes */
{
   Chromseglist *out = NULL;
   Chromseglist **curr_chrom, **tailaddr;
   Chromseglist *curr_first = x.maternal;
   Chromseglist *curr_second = x.paternal;
   Breaklist *curr_break = newbreaks;

   long long segbegin, segend;

   int   which;

   /* start tracking the list tail--begins as a NULL pointer */
   tailaddr = &out;

   segbegin = 0;
   /* randomly choose a chromosome to start on */
   which = rans () < 0.5;

   /* traverse the list of new breaks */
   while (curr_break)
   {
      /* current segment end is the breakpoint we're currently at */
      segend = curr_break->addr;
      /* change curr_chrom's pointee to be the current chromosome */
      curr_chrom = which ? &curr_first : &curr_second;
      /* traverse list until we're past the current segment start */
      while ((*curr_chrom)->current->endpos < segbegin)
      {
         *curr_chrom = (*curr_chrom)->next;
      }
      /* now traverse until we're past the current segment end,
       * but as we go put the breaks into the output chromosome
       * using add_chrom_tail */
      while ((*curr_chrom)->current->endpos < segend)
      {
         tailaddr =
            add_chrom_tail (tailaddr, (*curr_chrom)->current->fgl,
                            (*curr_chrom)->current->endpos);
         *curr_chrom = (*curr_chrom)->next;
      }
      /* finally, put the breakpoint in as a new segment */
      tailaddr = add_chrom_tail (tailaddr, (*curr_chrom)->current->fgl, segend);
      /* set new segment start to previous segment end */
      segbegin = curr_break->addr;

      curr_break = curr_break->next;
      /* switch current chromosome */
      which = (which + 1) % 2;
   }

   /* does the last bit after the last break */
   curr_chrom = which ? &curr_first : &curr_second;
   while ((*curr_chrom)->current->endpos < segbegin)
   {
      *curr_chrom = (*curr_chrom)->next;
   }
   while (*curr_chrom)
   {
      tailaddr =
         add_chrom_tail (tailaddr, (*curr_chrom)->current->fgl,
                         (*curr_chrom)->current->endpos);
      *curr_chrom = (*curr_chrom)->next;
   }
   /* eliminates redundant nodes in the output chromosome */
   weed_chrom (&out);

   return out;
}

void weed_chrom (Chromseglist ** x)
/* remove redundant junctions in the chromosome which is passed by reference.
 * makes things cleaner and makes finding IBD easier */
{
   Chromseglist *curr, *last;
   Chromseglist **curr_ptr;

   curr = (*x)->next;
   last = *x;
   curr_ptr = x;

   while (curr)
   {
      /* if fgl is the same, cut that node out of the list */
      if (curr->current->fgl == last->current->fgl)
      {
         (*curr_ptr) = curr;
         free (last->current);
         free (last);
      }
      else
         /* otherwise, advance the (pointer to pointer) */
         curr_ptr = &last->next;
      last = curr;
      curr = curr->next;
   }

   return;
}

Ibdlist *find_ibd (Chromseglist * first, Chromseglist * second)
/* the algorithm here is roughly:
 * - set a pointer to the first node of each chromosome
 * - set last_end == 0
 * 1. check for equality of fgl
 * 1. a. if yes, we have entered an IBD segment
 *       keep advancing both pointers
 *      until they are different (meaning we
 *      have left the IBD segment)
 *    b. the IBD seg starts at "last_end" and ends
 *      at the lesser of the two pointers
 * 2. set "last_end" to the endpos of the most recently
 *    moved pointer (pointed to by curr_ptr)
 * 3. advance the lesser of the two pointers
 * 4. go to 1
 *
 * The idea is to zig-zag up the pair of chromosomes
 * from this:
 * ______*______________________
 *      /
 * ____/________________________
 *    *
 * to this:
 * ______*______________________
 *        \
 * ________\____________________
 *          *
 *
 * comparing fgl at each step
 *
 * Seems to work in basic cases but need to make sure
 * it's working at beginning or end of chromosome
 */
{
   Ibdlist *out = NULL;
   Ibdlist **tail_ibd;
   Chromseglist *curr_first, *curr_second;
   Chromseglist **curr_ptr;
   long long last_end;

   tail_ibd = &out;
   curr_first = first;
   curr_second = second;
   curr_ptr = &curr_first;
   last_end = 0;

   while (*curr_ptr)
   {
      /* select smaller junction */
      if (curr_first->current->endpos < curr_second->current->endpos)
         curr_ptr = &curr_first;
      else
         curr_ptr = &curr_second;

      /* have we started an IBD segment? */
      if (curr_first->current->fgl == curr_second->current->fgl)
      {
         while (curr_first->next && curr_second->next &&
                curr_first->current->endpos == curr_second->current->endpos)
            /* advance through the IBD segment */
         {
            curr_first = curr_first->next;
            curr_second = curr_second->next;
         }

         /* IBD segment ends at smaller of first two unequal breaks */
         if (curr_first->current->endpos < curr_second->current->endpos)
            curr_ptr = &curr_first;
         else
            curr_ptr = &curr_second;

         tail_ibd =
            add_ibd_tail (tail_ibd, last_end, (*curr_ptr)->current->endpos);
      }

      /* advance the smaller junction */
      last_end = (*curr_ptr)->current->endpos;
      (*curr_ptr) = (*curr_ptr)->next;
   }

   return out;
}

/* Id number included just to tag output for later use.                       */
void printbreaks (Chromseglist * x)
{
   Chromseglist *curr = x;

   while (curr)
   {
      fprintf (oScorF, "%d %llu ", curr->current->fgl, curr->current->endpos);
      curr = curr->next;
   }
   fprintf (oScorF, "\n");

   return;
}

/* id number included just to tag output for later use */
void printibd (Ibdlist * x, int idnum)
{
   Ibdlist *curr = x;

   while (curr)
   {
      fprintf (oScorF, "%d\t%llu\t%llu\n", idnum, curr->start, curr->end);
      curr = curr->next;
   }

   return;
}

void printgeneration (Person * x, int length, int generation)
{
   int   i;

   for (i = 0; i < length; i++)
   {
      /* 4 id #s, sex, then 0 for maternal chrom, 1 for paternal */
      fprintf (oScorF, "%d %d %d %d %d 0 ", x[i].id, generation, x[i].motherid,
               x[i].fatherid, x[i].sex);
      printbreaks (x[i].maternal);
      fprintf (oScorF, "%d %d %d %d %d 1 ", x[i].id, generation, x[i].motherid,
               x[i].fatherid, x[i].sex);
      printbreaks (x[i].paternal);
   }

   return;
}

Person *initfounders (int n, long long chrlength, int startfgllab, int sex)
    /* sets up an array of length n of founders with chromosome length chrlength.
     * startfgllab is so we can start where a previous call to initfounders
     * left off (say, we already did the females and now we're doing males) */
{
   Person *out;

   int   fgllabel = startfgllab;
   int   i;

   out = (Person *) malloc (n * sizeof (Person));

   for (i = 0; i < n; i++)
   {
      out[i].id = labelcounter++;
      out[i].motherid = 0;
      out[i].fatherid = 0;
      out[i].sex = sex;

      /* founder chromosomes consist of just a single "breakpoint" at
       * the end of the chromosome and a single FGL */
      out[i].maternal = (Chromseglist *) malloc (sizeof (Chromseglist));
      out[i].maternal->current = (Chromseg *) malloc (sizeof (Chromseg));
      out[i].maternal->current->fgl = fgllabel++;
      out[i].maternal->current->endpos = chrlength;
      out[i].maternal->next = NULL;

      out[i].paternal = (Chromseglist *) malloc (sizeof (Chromseglist));
      out[i].paternal->current = (Chromseg *) malloc (sizeof (Chromseg));
      out[i].paternal->current->fgl = fgllabel++;
      out[i].paternal->current->endpos = chrlength;
      out[i].paternal->next = NULL;
   }

   return out;
}

void new_generation (Person * females, Person * males,
                     Person ** newfemales, Person ** newmales,
                     int nfemales, int nmales, int nkids, long long chrlength,
                     double rate, double remate_frac)
/* assume one couple per female in population
 * nkids is number of girls and number of boys
 * i.e. total kids / 2
 * newfemales and newmales are pointers to Persons passed as reference */
{
   double accept_prob;                  /* Pr( Accept | n prev marriages ) = remate_frac^n */

   int   random_mother;                 /* initalise things for rejection sampling */
   int   random_father;
   int   curr_mother_ind;
   int   curr_father_ind;
   int  *visits_mother;
   int  *visits_father;
   int   counter;
   int   i, j;

   Person curr_mother, curr_father;
   Person *curr_child;
   Breaklist *newbreaks;

   visits_mother = alloc_int1d (nfemales);
   visits_father = alloc_int1d (nfemales);

   (*newfemales) = malloc (nfemales * nkids * sizeof (Person));
   (*newmales) = malloc (nfemales * nkids * sizeof (Person));

   for (i = 0; i < nfemales; i++)
   {
    /**************************************************************************************/
      /*curr_mother = females[(int) floor (rans () * nfemales)]; */
      /*curr_father = males[(int) floor (rans () * nmales)];     */
    /*** Rejection Sampling for Mother *****/
      curr_mother_ind = -1;
      while (curr_mother_ind < 0)
      {
         random_mother = floor (rans () * nfemales); /* select a random mother */
         accept_prob = pow (remate_frac, (double) visits_mother[random_mother]); /* accept prob is remate_frac^visits, so is 1 if never visited */
         if (accept_prob > rans ())
         {
            curr_mother_ind = random_mother; /*update current mother */
            visits_mother[curr_mother_ind]++; /*increment number visits */
         }
         counter++;
      }
    /*** Rejection Sampling for Father *****/
      curr_father_ind = -1;
      while (curr_father_ind < 0)
      {
         random_father = floor (rans () * nfemales);
         accept_prob = pow (remate_frac, (double) visits_father[random_father]);
         if (accept_prob > rans ())
         {
            curr_father_ind = random_father;
            visits_father[curr_father_ind]++;
         }
      }
      /*set current mother and father values */
      curr_mother = females[curr_mother_ind];
      curr_father = males[curr_father_ind];

    /**************************************************************************************/
      for (j = 0; j < nkids; j++)
      {
         /* daughter */
         /* points to the element of the newfemales array we're currently on */
         curr_child = &(*newfemales)[nkids * i + j];
         /* give the kid an ID number */
         (*curr_child).id = labelcounter++;
         (*curr_child).motherid = curr_mother.id;
         (*curr_child).fatherid = curr_father.id;
         (*curr_child).sex = 0;
         /* create some breaks, use them to recombine parent's chromosomes,
          * then free the list of breaks (do this twice for each child */
         newbreaks = makebreaks (chrlength, rate);
         (*curr_child).maternal = recombine (curr_mother, newbreaks);
         free_breaks (newbreaks);
         newbreaks = makebreaks (chrlength, rate);
         (*curr_child).paternal = recombine (curr_father, newbreaks);
         free_breaks (newbreaks);

         /* son */
         /* points to the element of the newmales array we're currently on */
         curr_child = &(*newmales)[nkids * i + j];
         (*curr_child).id = labelcounter++;
         (*curr_child).motherid = curr_mother.id;
         (*curr_child).fatherid = curr_father.id;
         (*curr_child).sex = 1;
         newbreaks = makebreaks (chrlength, rate);
         (*curr_child).maternal = recombine (curr_mother, newbreaks);
         free_breaks (newbreaks);
         newbreaks = makebreaks (chrlength, rate);
         (*curr_child).paternal = recombine (curr_father, newbreaks);
         free_breaks (newbreaks);
      }
   }
   /*printf ( "Proposed females: %d Out of: %d \n", counter, nfemales ); */

   free_1d ((void *) &visits_mother);
   free_1d ((void *) &visits_father);

   return;
}

void free_breaks (Breaklist * list)
{
   Breaklist *curr, *prev;

   if (!list)
      return;
   prev = list;
   curr = list->next;
   while (curr)
   {
      free (prev);
      prev = curr;
      curr = curr->next;
   }
   free (prev);

   return;
}

void free_ibdlist (Ibdlist * list)
{
   Ibdlist *curr, *prev;

   if (!list)
      return;
   prev = list;
   curr = list->next;
   while (curr)
   {
      free (prev);
      prev = curr;
      curr = curr->next;
   }
   free (prev);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_chromseglist subroutine frees up the memory allocated for the     */
/* linked list of chromosome segments pointed to by the one argument to the   */
/* subroutine.                                                                */
/*                                                                            */
/******************************************************************************/

void free_chromseglist (Chromseglist * first_chromseg)
{
   Chromseglist *p_curr_seg;            /* Pointer to the current chromosome  */
   /*                                      segment element.                   */
   Chromseglist *p_next_seg;            /* Pointer to the next chromosome     */
   /*                                      segment element.                   */

   p_curr_seg = first_chromseg;
   while (p_curr_seg != NULL)
   {
      free (p_curr_seg->current);
      p_next_seg = p_curr_seg->next;
      free (p_curr_seg);
      p_curr_seg = p_next_seg;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_generation subroutine frees up the memory allocated for a single  */
/* generation of individuals.  The first argument passed to free_generation   */
/* is a Person pointer to the origin of the generation vector.  The second    */
/* argument is an integer providing the number of individuals allocated for   */
/* the generation vector.                                                     */
/*                                                                            */
/******************************************************************************/

void free_generation (Person * individuals, const int num_indivs)
{
   int   i1;                            /* Subscript to current individual.   */

   for (i1 = 0; i1 < num_indivs; ++i1)
   {
      free_chromseglist (individuals[i1].maternal);
      free_chromseglist (individuals[i1].paternal);
   }

   free_1d ((void *) &individuals);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The simpop_fgl subroutine was developed from simu.c and new_simu.c.        */
/*                                                                            */
/* This subroutine simulates patterns of identity by descent in a population  */
/* after a user specified number of generations.                              */
/*                                                                            */
/******************************************************************************/

#define PNAM "ibd_create"

int simpop_fgl (int argc, char **argv)
{
   FILE *oSeedF;

   Person *current_females;
   Person *current_males;
   Person *prev_females;
   Person *prev_males;

   double remate_frac;
   double rate;

   long long chrlength;

   int   gensize;
   int   nogens;
   int   printgens;

   int   nkids;
   int   g1;

   /* ................................  SET-UP   ............................ */

   setupBasic ("R", argc, argv, PNAM);

   /* Make local copies of the parameter statement values for simpop_fgl.     */

   remate_frac = RematWeight;
   rate = CMperMbp;
   chrlength = ChromLen_bp;

   gensize = NumFemsGen;
   nogens = NumOffGens;
   printgens = NumOutGens;

   /* Tell the user what scaling factor for the conversion of base pairs to   */
   /* centimorgans is being used in this run of simpop_fgl.                   */

   printf (" The scaling factor used during this run of simpop_fgl is");
   printf (" %7.4f cM per Mbp\n", rate);

   printf (" Remember to use this same scaling when fitting the output to a");
   printf (" marker map\n\n");

   /* Open the output scores file to use as simpop_fgl's results file.        */

   if (oScorF == NULL)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   /* Initialize a few variables.                                             */

   labelcounter = 1;
   nkids = 1;

   current_females = initfounders (gensize, chrlength, 1, 0);
   current_males = initfounders (gensize, chrlength, 2 * gensize + 1, 1);

   if (nogens == printgens - 1)
   {
      printgeneration (current_females, gensize, 0);
      printgeneration (current_males, gensize, 0);
   }

   for (g1 = 0; g1 < nogens; ++g1)
   {
      if (g1 > 0)
      {
         free_generation (prev_females, gensize / nkids);
         free_generation (prev_males, gensize / nkids);
      }

      prev_females = current_females;
      prev_males = current_males;

      new_generation (prev_females, prev_males, &current_females,
                      &current_males, gensize, gensize, nkids, chrlength, rate,
                      remate_frac);

      gensize = gensize * nkids;

      if (nogens - g1 <= printgens)
      {
         printgeneration (current_females, gensize, g1 + 1);
         printgeneration (current_males, gensize, g1 + 1);
      }
   }

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   if (oScorF)
      fclose (oScorF);

   /* We are done, so free up the remaining storage that has been allocated   */
   /* by the simpop_fgl subroutine.                                           */

   free_generation (prev_females, gensize / nkids);
   free_generation (prev_males, gensize / nkids);

   free_generation (current_females, gensize / nkids);
   free_generation (current_males, gensize / nkids);

   free_fnams ();

   return 0;
}
