/******************************************************************************/
/*                                                                            */
/* The compute_ibd_label subroutine returns an integer containing the label   */
/* of the IBD state passed in an integer vector as the second argument.  The  */
/* first argument, state_id_length, is the number of integers in the second   */
/* argument vector.                                                           */
/*                                                                            */
/******************************************************************************/

int compute_ibd_label (int state_id_length, int *state_id)
{
   int   label;                         /* Local copy of the IBD state label. */
   int   c1;                            /* Subscript in the state_id vector.  */

   label = 0;

   for (c1 = 0; c1 < state_id_length; ++c1)
      label = (label * c1) + (state_id[c1] - 1);

   return label;
}

/******************************************************************************/
/*                                                                            */
/* The compute_ibd_probs subroutine is called to compute estimates of the     */
/* conditional probability that genes are shared IBD for all score sets.      */
/*                                                                            */
/* If the user has included a "set proband gametes all individual pairs"      */
/* parameter statement then there is only 1 score set consisting of 4 gametes */
/* belonging to the pair of individuals identified by the indiv1 and indiv2   */
/* subscripts passed as the first 2 arguments.  Otherwise, if the this is not */
/* a DoingAllPairs case, then there are as many score sets as there are "set  */
/* proband gametes ..." parameter statements with distinct score set numbers  */
/* and the indiv1 and indiv2 arguments are not used by the subroutine.        */
/*                                                                            */
/******************************************************************************/

void compute_ibd_probs (int indiv1, int indiv2, FILE * oScorF, FILE * oXtra)
{
   double **qq_ibd_initial;             /* Initial IBD probabilities.         */
   double **qq_ibd_intermediate;        /* Intermediate IBD probabilities.    */
   double **qq_ibd_final;               /* Final IBD probabilities.           */

   double **pprob;

   char **state_names = NULL;           /* Possible gene IBD states names.    */
   char *state_id;                      /* Returned state identification.     */

   int  *simStates;                     /* A vector containing the subscripts */
   /*                                      to each marker's simulated state.  */

   int **haplodat;                      /* Haplotypic data.                   */
   int  *hapdat;

   int   ns_final;                      /* Final number of gene IBD states.   */

   int   indiv;                         /* The haplodat subscript to the      */
   /*                                      current individual.                */
   int   ix;                            /* The phasedatindiv subscript to the */
   /*                                      current individual.                */

   int   tmp1, tmp2;                    /* The two tmp variables are holders  */
   /*                                      for swapping alleles.              */

   int   ss_size;                       /* Size of the current score set.     */
   int   ss_size_ndx;                   /* Current score set size subscript.  */
   int   ss_indiv_cnt;                  /* The number of individuals which    */
   /*                                      make up the score set.             */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Subscript to current score set.    */

   int   g1;                            /* Subscript to proband gamete.       */
   int   m1;                            /* Subscript to current marker.       */
   int   s2;                            /* Subscript to current IBD state.    */

   int   r1;                            /* Subscript for current realization. */
   int   state_ndx;                     /* Subscript to the current marker's  */
   /*                                      simulated state.                   */
   int   state_label;                   /* Simulated IBD state's label.       */

   /* In the main loop for the ibd_haplo progam compute the IBD probabilities */
   /* for each score set indicated by the ibd_haplo user.                     */

   for (s1 = 0; s1 < NScSet[c1]; ++s1)  /* For each score set:                */
   {
      ss_size = NPrGam[c1][s1];         /* Get size of the current score set. */

      printf (" === Start of the computation of IBD probabilities for score");
      printf (" set %*d ===\n", nsn, ScSetNo[c1][s1]);
      fflush (stdout);

      haplodat = alloc_int2d (ss_size, NMarks);
      hapdat = alloc_int1d (NMarks);

      /* Copy the necessary genotypic data from the ObsMark[][][] array.      */

      for (g1 = 0; g1 < ss_size; ++g1)
      {
         ix = ProbIx[c1][s1][g1];

         for (m1 = 0; m1 < NMarks; ++m1)
            haplodat[g1][m1] = ObsMark[ix][m1][ProbGam[c1][s1][g1]];
      }

      phasdat = NULL;                   /* Default to no phasing data.        */

      /* Make the changes needed in haplodat to accomodate the partially      */
      /* phased data case.                                                    */

      if (PhaseofData == PARTIALPHASE)
      {
         /* When using phased data, set the relevant pointer.                 */
         phasdat = alloc_int1d (NMarks);

         /* There are 2 gametes for each individual in the score set.         */
         ss_indiv_cnt = ss_size / 2;

         for (indiv = 0; indiv < ss_indiv_cnt; ++indiv)
         {
            ix = (s1 * 2) + indiv;

            for (m1 = 0; m1 < NMarks; ++m1)
            {
               if (phasdatindiv[ix][m1] == DISCORDANT)
               {
                  tmp1 = haplodat[2 * indiv][m1];
                  tmp2 = haplodat[2 * indiv + 1][m1];
                  haplodat[2 * indiv][m1] = tmp2;
                  haplodat[2 * indiv + 1][m1] = tmp1;
                  phasdatindiv[ix][m1] = CONCORDANT;
               }

               /* It's only phased for the HMM model if BOTH gametes for the  */
               /* individual are phased.                                      */
               phasdat[m1] = phasdatindiv[2 * s1][m1] *
                  phasdatindiv[2 * s1 + 1][m1];
            }
         }

         if (DebugMain)
            print_phase_data (s1, nc);
      }

      /* Restore the data structures needed for this score set's size.        */

      ss_size_ndx = ss_size - 1;        /* Set the score set size subscript.  */

      state_matrix = score_set_size_pointers[ss_size_ndx].state_matrix;
      num_orbits = score_set_size_pointers[ss_size_ndx].num_orbits;
      orbitSizes = score_set_size_pointers[ss_size_ndx].orbitSizes;
      orbits = score_set_size_pointers[ss_size_ndx].orbits;
      marg_probs = score_set_size_pointers[ss_size_ndx].marg_probs;
      tmat_rate = score_set_size_pointers[ss_size_ndx].tmat_rate;
      tmat = score_set_size_pointers[ss_size_ndx].tmat;
      transmat = score_set_size_pointers[ss_size_ndx].transmat;

      /* Get the number of haplotypic states for the score set.               */
      num_states = PTgg[ss_size];

      if (DebugLevel == 4)
      {
         fprintf (stdout, " Markers = %d,", NMarks);
         fprintf (stdout, " proband gametes = %d,", ss_size);
         fprintf (stdout, " haplotypic states = %d\n", num_states);
         fflush (stdout);
      }

      /* Determine what the final number of gene IBD states output will be.   */

      if (PhaseofData == PARTIALPHASE || PhaseofData == UNPHASED)
         ns_final = compute_number_orbits ();
      else
         ns_final = num_states;

      /* If the user has named an output extra file, then output the score    */
      /* set identification information to the output extra file, 1 line per  */
      /* score set.                                                           */

      if (oXtra)
      {
         fprintf (oXtra, " %*d", nsn, ScSetNo[c1][s1]);
         fprintf (oXtra, " %2d", NPrGam[c1][s1]);
         fprintf (oXtra, " %*d", nst, ns_final);

         for (g1 = 0; g1 < ss_size; ++g1)
         {
            fprintf (oXtra, "  %*s", nc, ProbNam[c1][s1][g1]);
            fprintf (oXtra, " %d", ProbGam[c1][s1][g1]);
         }

         fprintf (oXtra, "\n");
         fflush (oXtra);
      }

      /* Call make_data to encode the alleles found for each gamete in the    */
      /* score set into a single integer variable for each marker.            */

      make_data (hapdat, haplodat, ss_size, 0, NMarks, 1, 0);

      pprob = alloc_dbl2d (NMarks, num_states);
      qq_ibd_initial = alloc_dbl2d (NMarks, num_states);

      hap_fwd_compute (hapdat, NMarks, num_states, allefreq, marg_probs,
                       Mdist, pprob, PhaseofData, phasdat, ss_size);

      if (NRealizations > 0)
      {
         simStates = alloc_int1d (NMarks);

         for (r1 = 0; r1 < NRealizations; ++r1)
         {
            /* For each realization do the following:                         */

            hap_sample (NMarks, num_states, simStates, marg_probs, Mdist,
                        pprob, PhaseofData, ss_size);

            for (m1 = 0; m1 < NMarks; ++m1)
            {
               state_ndx = simStates[m1];
               state_label = compute_ibd_label (ss_size,
                                                state_matrix[state_ndx]);
               state_id = compute_ibd_state (state_ndx, ss_size);

               fprintf (oScorF, " %*d ", nsn, ScSetNo[c1][s1]);
               fprintf (oScorF, " %*d ", nm, MarSel[m1]);
               fprintf (oScorF, " %10.6f ", snppos[m1]);
               fprintf (oScorF, " %*d ", nst, state_ndx + 1);
               fprintf (oScorF, " %*d ", nst, state_label);

               if (DebugLevel == 0)
               {
                  for (g1 = 0; g1 < ss_size; ++g1)
                     fprintf (oScorF, "%2d", state_matrix[state_ndx][g1]);
               }
               else
                  fprintf (oScorF, " %s", state_id);

               fprintf (oScorF, "\n");

               free_1d ((void *) &state_id);
            }
         }

         free_1d ((void *) &simStates);
      }

      else if (NRealizations == 0)
      {
         hap_bck_compute (hapdat, NMarks, num_states, allefreq, marg_probs,
                          Mdist, pprob, PhaseofData, phasdat, ss_size,
                          qq_ibd_initial);

         if (DebugLevel == 2)
         {
            fprintf (oScorF, " The qq_ibd_initial matrix returned from");
            fprintf (oScorF, " hap_bck_compute is:\n");

            haplo_debug_qq_ibd (nm, num_states, qq_ibd_initial, oScorF);
         }
      }

      if (PhaseofData == PARTIALPHASE || PhaseofData == UNPHASED)
      {
         /* Call the subroutine to collapse the haplotypic states down to the */
         /* distinct genotypic states.                                        */

         collapse_to_distinct (qq_ibd_initial, ns_final, &qq_ibd_intermediate,
                               ss_size, &state_names);
         free_2d ((void *) &qq_ibd_initial);
      }
      else
         qq_ibd_intermediate = qq_ibd_initial;

      if (DebugLevel == 2)
      {
         fprintf (oScorF, " The qq_ibd_intermediate matrix returned from");
         fprintf (oScorF, " collapse_to_distinct is:\n");

         haplo_debug_qq_ibd (nm, ns_final, qq_ibd_intermediate, oScorF);
      }

      /* When lexicographical output order is not wanted and if the number of */
      /* haplotypes in the current score set is 4, we still want to reorder   */
      /* the lexicographically ordered qq_ibd_intermediate entries into the   */
      /* "traditional" output order previously used.                          */

      if (Lexico == N && ss_size == 4 && DebugLevel != 3)
      {
         qq_ibd_final = alloc_dbl2d (NMarks, ns_final);
         if (PhaseofData == PARTIALPHASE || PhaseofData == UNPHASED)
         {
            for (m1 = 0; m1 < NMarks; ++m1)
               for (s2 = 0; s2 < ns_final; ++s2)
                  qq_ibd_final[m1][s2] =
                     qq_ibd_intermediate[m1][lex2gen_un[s2]];

         }
         else
         {
            for (m1 = 0; m1 < NMarks; ++m1)
               for (s2 = 0; s2 < ns_final; ++s2)
                  qq_ibd_final[m1][s2] =
                     qq_ibd_intermediate[m1][lex2gen_ph[s2]];
         }
         free_2d ((void *) &qq_ibd_intermediate);
      }
      else
         qq_ibd_final = qq_ibd_intermediate;

      if (NRealizations == 0)
      {
         if (DebugLevel == 3)
         {
            fprintf (oScorF, " The qq_ibd_final matrix is:\n");
            if (PhaseofData == PHASED)
               haplo_debug_phased (nm, ns_final, qq_ibd_final, oScorF);
            else
               haplo_debug_unphased (nm, ns_final, qq_ibd_final, state_names,
                                     oScorF);
            fprintf (oScorF, "\n");
         }
         else
         {
            for (m1 = 0; m1 < NMarks; ++m1)
            {
               fprintf (oScorF, " %*d ", nsn, ScSetNo[c1][s1]);
               fprintf (oScorF, " %*d ", nm, MarSel[m1]);
               fprintf (oScorF, " %10.6f ", snppos[m1]);
               for (s2 = 0; s2 < ns_final; ++s2)
                  fprintf (oScorF, " %7.4f", qq_ibd_final[m1][s2]);
               fprintf (oScorF, "\n");
            }
         }
      }

      fprintf (stdout, "\n");
      if (s1 < NScSet[c1] - 1)          /* Unless this is the last score set, */
         fprintf (stdout, "\n");        /* print another blank line.          */
      fflush (stdout);

      if (PhaseofData == PARTIALPHASE)
         free_1d ((void *) &phasdat);

      if (qq_ibd_final)
         free_2d ((void *) &qq_ibd_final);

      free_2d ((void *) &haplodat);
      free_1d ((void *) &hapdat);
      free_2d ((void *) &pprob);

      free_2d ((void *) &state_names);
   }

   return;
}
