/******************************************************************************/
/*                                                                            */
/*  Program name: gl_lods.c (Estimate or calculate LodScore by component,     */
/*                           using the results of gl_auto).                   */
/*                                                                            */
/*  EAT, April 2010.                                                          */
/*  SML, June  2012,  December 2012,  March 2013,  August 2013.               */
/*                                                                            */
/*  Note that this program is modified from lm_linkage.c in Lodscore.         */
/*                                                                            */
/*  Also note it will not do MCMC, but instead will take gl_auto output to    */
/*    compute lod score contributions at the selected markers.                */
/*                                                                            */
/******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "gl_lods"
#define GLLODS

#include "lodscore.h"

#include "gl_lods_rtnes.c"
#include "lods_debug.c"

int main (int argc, char *argv[])
{
   FILTAB_PTR ptb;                      /* File table pointer of OSCORE file. */
   FILE *fp2;                           /* Current component IBDgraphs file.  */

   IBDGraphList *ibd_graphs;            /* Linked list of IBD graphs.         */
   IBDGraphLocationEquivalences *igeq;  /* Pointer to structure containing    */
   /*                                      the location_lists.                */

   long  num_equiv_classes;             /* Number of equivalence classes.     */
   long  c1;                            /* Equivalence class subscript.       */

   int  *comp_num_indiv;                /* Number of individuals copied for a */
   /*                                      component's reduced IBDgraph file. */

   int   comp_ndx;                      /* Current component subscript.       */
   int   comp_no;                       /* Current component number.          */
   int   nghd_comp_no;                  /* Component number found in nghd[]   */
   /*                                      for the IBDgraphs to be copied to  */
   /*                                      the reduced file.                  */

   /* ....................................................................... */

   setupwithIndivList (argc, argv, PNAM, 0);
   setSampling ();

   /* At this point we have:
    *    PedSize -- original meaning of number read from pedigree file:
    *       for whom the data (possibly missing) has been stored in nghds.
    *    NMarks which is the number of markers at which lod scores are to
    *       be evaluated.
    *    MarLod which is a list of these markers by index in gl_auto output:
    *      Note: gl_auto is modified so that both programs can use
    *       original input marker indexing.
    *      Note: if a marker is called for that is not in the gl_auto output
    *       reconstruction of FGL labels will result in the evaluation at
    *       the proximal preceding marker that is in the gl_auto output.
    *    NIBDgraphs is the maximum number of scored FGL graphs to be processed
    *       for any component from the gl_auto output.
    *   Modified March 2011: EAT; number of individuals in gl_auto file will
    *     be determined on read-in; it is no longer provided by PedSize.      */

   if (DebugLevel == 2)
      lods_prnt_trait_data ();

   printf ("\n");
   fflush (stdout);

   /* The following comments were applicable to gl_lods before the code to    */
   /* support multiple components was added.  The same tasks need to be done, */
   /* but the location in the program where they are done has now been split  */
   /* between the calls to find_num_indivs and find_max_fgl() made here and   */
   /* the call to get_component_graphs() made below for each component.       */

   /* First we will reduce the DGL graphs to observed individuals.            */
   /* They are input in "extra file" and will be output to "output scores".   */
   /* We will then close output scores, and re-read it for computations.      */

   /* We will reduce the DGL graph to the set of observed individuals, and    */
   /* output it by nghd index, not by name.  (Note the index is 0-origin.)    */

   /* Names in DGL but not in nghd will be dropped.                           */
   /* Names in DGL and in nghd but with nghd.obs == 0 will be dropped.        */
   /* Names NOT in DGL but in nghd, with ngdh.obs != 0 will be WARNED.        */

   find_num_indivs (&tot_num_indivs);   /* Find total number of individuals.  */
   printf (" Number of individuals in IBD graphs file is %d\n", tot_num_indivs);
   use_num_graphs = min (input_num_graphs, (long) NIBDgraphs);

   /* Allocate storage for the DGL_names and DGL_index vectors.               */

   DGL_names = alloc_char2d (tot_num_indivs, MAXNAMLEN + 1);

   DGL_index = (int *) malloc (tot_num_indivs * sizeof (int));
   alloc_chk (WHERE, DGL_index);
   memset (DGL_index, INVALID_INDEX, tot_num_indivs * sizeof (int));

   if (input_num_graphs < NIBDgraphs)
      printf (" %d graphs were requested, but only %ld were given (W)\n",
              NIBDgraphs, input_num_graphs);
   printf ("\n");
   fflush (stdout);

   /* Allocate penetrances, and set Normal constants for quantitative traits. */

   if (TraitTyp == QUANTITATIVE)
      set_qtlpen (loci[0].ve);

   find_max_fgl (&nFGL);                /* Find maximum FGL used.             */
   alloc_gen_pen (nFGL, NLoci);

   printf (" Number of individuals in nghd structure is %d\n", pedsize);
   printf (" Have alloced gen_pen nFGL=%d, NLoci=%d\n\n\n", nFGL, NLoci);
   fflush (stdout);

   /* The IBDgraphs file is now in a format that the IBDgraph software can    */
   /* use.  So we have a smaller set to run, with added weights.              */

   /* Allocate an integer vector to store the number of individuals in each   */
   /* component's reduced IBD graphs file.                                    */

   comp_num_indiv = alloc_int1d (num_comp_input);

   for (comp_ndx = 0; comp_ndx < num_comp_input; ++comp_ndx)
   {
      /* Create a temporary file made up of only those IBD graphs belonging   */
      /* to the current component.                                            */

      comp_no = comp_ndx + 1;
      nghd_comp_no = get_component_graphs (comp_no, &comp_num_indiv[comp_ndx]);
      printf (" ====== Computing LodScore for component %d ======\n\n",
              nghd_comp_no);

      printf (" Number of observed individuals in each IBD graph for");
      printf (" component %d is %d\n", nghd_comp_no, comp_num_indiv[comp_ndx]);

      /* If there are no individuals with IBD graphs in the current component */
      /* there is nothing to compute, so continue with the next component.    */

      if (comp_num_indiv[comp_ndx] == 0)
         continue;

      /* Reopen the temporary file just created to be reread by the IBDgraph  */
      /* library subroutines.                                                 */

      ptb = get_filent (OSCORE);        /* Reopen reduced IBD graphs file.    */
      if (!(fp2 = fopen (ptb->fname, "r")))
         give_up (ERR_qit);
      if (DebugLevel == 6)
         printf (" Reopened component %d's reduced IBD graphs file \"%s\"\n\n",
                 nghd_comp_no, ptb->fname);
      fflush (stdout);

      ibd_graphs = NewIBDGraphList ();  /* Initializing the graph list.       */
      createIBDGraphs (ptb->fname, ibd_graphs); /* Build the graphs.          */

      /* Allocate the storage needed for the arrays required for this         */
      /* component.  Note that the entries in these arrays are not            */
      /* initialized.                                                         */

      gl_lods_alloc (use_num_graphs);

      /* Determine the equivalence_classes for each selected marker.          */

      det_equivalence_classes (ibd_graphs);

      /* Initialize the IBDGraphLocationEquivalences structure.  Obtain the   */
      /* total number of equivalence classes allocated in this structure.     */

      igeq = NewIBDGraphLocationEquivalences (ibd_graphs);
      num_equiv_classes = IBDLocEq_NumClasses (igeq);

      printf (" Grouped %ld locations into %ld equivalence classes\n",
              IBDLocEq_TotalSize (igeq), num_equiv_classes);

      if (DebugLevel == 5)
      {
         printf ("\n");
         IBDGraphLocationEquivalences_Print (igeq);
      }
      printf ("\n");

      /* For each equivalence class, the lodscore_elements for markers within */
      /* the equivalence class are collapsed to one single lodscore_element.  */

      for (c1 = 1; c1 < num_equiv_classes; ++c1)
         collapse_lodscore_elements (igeq, c1);

      O_DecRef (igeq);

      /* Now that the pointer constructs are complete, log likehoods for each */
      /* lodscore_element need to be computed.                                */

      prnt_contributions (fp2, use_num_graphs, comp_num_indiv[comp_ndx],
                          nghd_comp_no);

      O_DecRef (ibd_graphs);

      /* If the user has provided each component's base log likelihood,       */
      /* gl_lods can compute component lod scores for the selected markers.   */

      gl_lods_loglikes (nghd_comp_no);

      if ((num_comp_input - comp_ndx) > 1)
         printf ("\n");
      fflush (stdout);

      /* Free up the storage used by gl_lods for the current component.       */

      gl_lods_free ();

      /* Close the reduced IBD graphs file.                                   */

      fclose (fp2);
   }

   free_1d ((void *) &comp_num_indiv);
   free_DGL_vectors ();

   exit (EXIT_SUCCESS);
}
