/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/**
 * This function builds global data structures used by the Lodscore programs.
 *
 * proc_lods() has as its input the results of parsing the parameter statements
 * for the LR_Lods programs: lm_lods and lm_schnell, and for the Lodscore
 * programs: lm_bayes, lm_linkage and lm_twoqtl.
 * proc_lods() uses many of the same data structures and functions as used
 * by proc_drop() and proc_auto().
 * This function checks parameter statement data for consistency and complete-
 * ness and, if there are no errors, builds the global data structures for the
 * Morgan main programs and subroutine libraries.
 *
 * One chromosome only is selected.
 * A set of test positions is assigned for one or two trait loci (tlocs).
 *
 * Any combination of sex-not-specified maps or sex-specific maps may be used.
 *
 *      MAJ  revs  5/2000, 1/2002, 8/2003, 6/2005, 6/2006, 10/2007.
 *      SML  revs  9/2008, 8/2009.
 */

int proc_lods (void)
{
   int   entry, sex;
   int   m_no, tl_no, t_no, loc_no, x1;
   int   sxsp_maps = FALSE;             /* Whether or not sex-spec maps used. */
   SIM_PTR p_s;
   TRAITTLOC_PTR p_tt;
   PRNTMAP_PTR p_m;
   PRDIST_PTR p_l;
   VMAP_PTR p_vmap;
   PSMAP_PTR p_si;

   if (!test_lods_sel ())               /* IMPORTANT: proc_lods expects one   */
      if (proc_err ())                  /* trait with varying test positions  */
         return FALSE;                  /* for one or two trait loci (tlocs). */

   if (n_err)
      return proc_end ();

   if (TRAIT_SELCTD && TRAIT_TLOC)      /* Finds the tloc(s) for the selected */
      if (!find_trait_tloc ())          /* trait.                             */
         if (proc_err ())
            return FALSE;

   if (n_err)
      return proc_end ();

   if (!find_tr_tloc_vmap ())           /* Finds the chromosome map(s) for    */
      if (proc_err ())                  /* test positions for the tloc(s).    */
         return FALSE;

   if (n_err)
      return proc_end ();

   entry = FALSE;
   p_s = nxt_sim_chrom (&entry);        /* One chromosome only is selected.   */

   if (!updt_sim_vtloc (p_s))           /* Add tloc(s) to chromosome selected. */
      if (proc_err ())
         return FALSE;

   if (MARK_MAP)
   {
      if (!test_mark_map (p_s, &sxsp_maps)) /* Checks the marker map.         */
         if (proc_err ())
            return FALSE;
   }

/* Check on the test positions for all selected tlocs in the trait_tloc list. */

   entry = FALSE;
   while ((p_tt = nxt_tr_tl (&entry)))
   {
      if (p_tt->selectd)
      {
         for (x1 = 0; x1 < p_tt->no_loci; ++x1)
         {
            tl_no = p_tt->p_tloc[x1];
            if (!test_tloc_vmap (p_s, tl_no, &sxsp_maps))
               if (proc_err ())
                  return FALSE;
            if (!test_tloc_vext (p_s, tl_no))
               if (proc_err ())
                  return FALSE;
            if (!test_tloc_vint (p_s, tl_no))
               if (proc_err ())
                  return FALSE;
         }
      }
   }

   entry = FALSE;
   while ((tl_no = nxt_sim_mapd_tl_no (&entry, p_s)))
      mak_tloc_list (tl_no, LNKD_TLOC); /* Adds required tloc to the list.    */

   if (MARK_ALLELE)                     /* For each marker, check allele      */
   {                                    /* frequencies.                       */
      if (!test_locus_nodes (p_s))
         if (proc_err ())
            return FALSE;

      entry = FALSE;
      while ((m_no = nxt_sim_mark_no (&entry, p_s)))
         if (!test_mark_freq (p_s, m_no))
            if (proc_err ())
               return FALSE;
   }

   if (TRAIT_SELCTD && TLOC_ALLELE)     /* For each tloc, check allele        */
   {                                    /* frequencies.                       */
      entry = FALSE;
      while ((tl_no = nxt_sim_tl_no (&entry)))
         if (!test_tloc_freq (tl_no))
            if (proc_err ())
               return FALSE;
   }

   entry = FALSE;
   t_no = nxt_sim_tr_no (&entry);       /* Get the trait number for the tloc. */

   if (TRAIT_SELCTD)
   {
      set_trait_typ (t_no);             /* Establish what type of trait is    */
      if (!test_trait_typ ())           /* selected.                          */
         if (proc_err ())
            return FALSE;
   }

   if (TraitTyp == QUANTITATIVE)
   {
      if (sim_traits && TRT_TLC_GENO)
      {
         entry = FALSE;
         while ((tl_no = nxt_tr_tl_no (&entry, t_no)))
            if (!test_tr_tl_genotyp (t_no, tl_no)) /* Check genotype means.   */
               if (proc_err ())
                  return FALSE;
      }

      if (sim_traits && TRAIT_VARI)
         if (!test_trait_vars (t_no))   /* Check additive and residual        */
            if (proc_err ())            /* variances.                         */
               return FALSE;

/* For a quantitative trait, find the selected trait_tloc_node from a "set    */
/* trait K1 multiple tlocs L1 L2 ..." statement to determine how many trait   */
/* loci have been specified by the user.  Save this number as NQtl.           */

      entry = FALSE;
      while ((p_tt = nxt_tr_tl (&entry)))
      {
         if (p_tt->trait_no == t_no && p_tt->selectd)
            NQtl = p_tt->no_loci;
      }

   }

   if (n_err)
      return proc_end ();

   if (MARK_MAP)                        /* Makes the "print map" for the sel  */
   {                                    /* markers and tloc test positions.
                                         * Assigns loci to sex-specific maps.
                                         * For maps not present, defaults to
                                         * AVG, then F or M maps.             */
      prnt_chrom_head (p_s);

      for (sex = FEM; sex <= AVG; ++sex) /* For each gender,                  */
      {
         p_m = mak_markprmap (p_s, sex); /* First, for the selected markers,  */

         if (MARK_MIN_DIST)             /* Checks distances between markers,  */
            if (!chk_markmindist (p_m)) /* if required.                       */
               if (proc_err ())
                  return FALSE;

         if (!mak_tlocvprmap (p_s, p_m, sex)) /* Checks test positions of the */
            if (proc_err ())            /* selected tloc(s).                  */
               return FALSE;

         shrink_prntmap (p_m);          /* Removes non selected marker loci.  */
         fix_prmap_dist (p_m);          /* Finds distances between adjacent   */
         /*                                markers for printing.              */
      }

      if (n_err)
         return proc_end ();

      prnt_map (sxsp_maps);             /* Prints the chromosome map.         */

      /* Print a tloc map for each selected tloc.                             */
      entry = FALSE;                    /* Use default rules for gender.      */
      while ((p_vmap = nxt_tl_vmap (&entry, AVG, p_s)))
      {
         if (p_vmap->selectd)
            prnt_vmap (p_vmap, p_m->no_mark_sel, sxsp_maps);
      }

      if (PrintLevel >= 4)
         fprintf (stdout, "\n");

      if (MARK_ALLELE)                  /* Prints marker allele frequencies.  */
         prnt_mark_freqs (p_s);
   }

   if (sim_traits && TLOC_ALLELE)       /* Prints tloc allele frequencies.    */
      prnt_vtloc_freqs (p_s);

   if (TRAIT_SELCTD && TraitTyp == QUANTITATIVE)
   {                                    /* List traits and prints parameters. */
      prnt_trait_head ();

      if (TRT_TLC_GENO)
         prnt_tr_tl_genotyp ();
      else if (TRAIT_TLOC)
         prnt_tr_tl_list ();
      else if (!TRAIT_VARI)
         prnt_tr_list ();

      if (TRAIT_VARI)
         prnt_trait_vari ();

      if (PrintLevel >= 2)
         fprintf (stdout, "\n");
   }

   NMarks = p_m->no_mark_sel;           /* Counts loci.                       */
   NLnkdTL = p_m->no_tloc_sel;
   if (UNLTLOC_SELCTD)
      NUnlnTL = sim_unlnkd_tlocs->no_tloc_sel;
   else
      NUnlnTL = 0;

   NTLocs = NLnkdTL + NUnlnTL;
   NLoci = NMarks + NTLocs;             /* Total NLoci includes the tlocs.    */

   if (!sim_traits)
      NTraits = 0;
   else
      NTraits = sim_traits->no_traits;

   entry = FALSE;
   while ((tl_no = nxt_sim_tl_no (&entry)))
      mak_tlocnam (tl_no);              /* Add this tloc name.                */

   alloc_marek ();                      /* Stores marker recombination        */
   mak_marek ();                        /* fractions in Marek[][].            */
   mak_tarek ();                        /* Stores tloc recomb fract and pos   */
   /*                                      (or .5, .5 if no linked tloc).     */

   /* Construct the locpos data for each selected tloc.                       */
   entry = FALSE;                       /* Use default rules for gender.      */
   while ((p_vmap = nxt_tl_vmap (&entry, AVG, p_s)))
   {
      if (p_vmap->selectd)
      {
         /* If not previously determined, calculate NTPos, the number of      */
         /* varying test tloc positions.  Also initialize NTPos1.             */
         if (NTPos == UNDEF)
         {
            NTPos = 0;
            p_si = p_vmap->p_vprntmap[FEM]->p_sintv;
            while (p_si)
            {
               NTPos += p_si->no_pos;
               p_si = p_si->p_next;
            }
            NTPos1 = NTPos + 1;

            alloc_intval ();            /* Sets positions in IntVal[]         */
            mak_intval (p_vmap);        /* and recomb fracts in RecVal[][]    */
            mak_recval (p_vmap);        /* (arrays are 2 greater than NTPos). */

            mak_locpos (p_s, p_vmap);   /* Stores info about all loci.        */
         }
      }
   }

   entry = FALSE;                       /* For each locus, stores allele      */
   p_m = nxt_prntmap (&entry, AVG);     /* frequencies in Qq[][].             */

   if (MARK_ALLELE || TLOC_ALLELE)      /* First markers and linked tloc      */
   {
      entry = FALSE;
      while ((p_l = nxt_prntmap_locus (&entry, p_m)))
         if (p_l->locus_typ == TLOC_LOC && INHER_SELCTD)
            mak_allefrq (NULL, NULL, 0);
         else
            mak_allefrq (p_m, p_l, p_l->no);

      if (UNLTLOC_SELCTD)               /* or unlinked tloc,                  */
      {
         entry = FALSE;
         while ((tl_no = nxt_sim_unln_tl_no (&entry)))
            mak_allefrq (NULL, NULL, tl_no);
      }

      else if (NTPos > 0)               /* or if test positions map used,     */
      {
         entry = FALSE;
         while ((p_vmap = nxt_tl_vmap (&entry, AVG, p_s)))
            if (p_vmap->selectd)
               mak_allefrq (NULL, NULL, p_vmap->tlno);
      }

      else if (!NTLocs)                 /* or if there is no tloc.            */
         mak_allefrq (NULL, NULL, 0);
   }

   if (SEL_MARK)
   {
      alloc_marsel ();
      loc_no = 0;

      entry = FALSE;
      while ((p_l = nxt_prntmap_locus (&entry, p_m)))
      {
         if (p_l->locus_typ == MARK_LOC)
         {
            MarSel[loc_no] = p_l->no;
            ++loc_no;
         }
      }

      if (MARK_MAP)
         mak_mdist (p_s);
   }

   if (MARK_NAME)
   {
      if (!test_marknam (p_s))          /* Checks marker names.               */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         if (mak_marknam (p_s))         /* Stores and prints marker names,    */
            prnt_marknam (p_s);         /* if provided.                       */
      }
      else
         return proc_end ();
   }

   if (MARK_DATA)
   {
      if (!test_markdat (p_s))          /* Checks marker data.                */
         if (proc_err ())
            return FALSE;

      if (!n_err)
      {
         mak_markdat (p_s);             /* Stores and prints marker data.     */
         prnt_markdat (p_s);
      }
      else
         return proc_end ();
   }

   if (LM_BAYES)
   {
      if (!test_ps_pr ())               /* Checks pseudo-priors.              */
         if (proc_err ())
            return FALSE;
   }

   if (!test_lods_pars ())              /* Check MC computational parameters. */
      if (proc_err ())
         return FALSE;

/* Stores incomplete penetrances, genotype means and variances.               */
   if (NTraits && (TraitTyp == QUANTITATIVE || TraitTyp == DISCRETE))
   {
      alloc_trait_pars (NTraits);
      entry = FALSE;
      t_no = nxt_sim_tr_no (&entry);    /* Get the trait number.              */
      pass_trait (t_no, 0);
   }

   if (TRAIT_DATA)                      /* Checks input record parameters.    */
      if (!test_ped_trtdat ())
         if (proc_err ())
            return FALSE;
   if (n_err)
      return proc_end ();

   if (pmx->max_ped > 0)
      mak_ped_dat ();

   if (TRAIT_DATA)                      /* Sets up and reports trait pedigree */
   {                                    /* data.                              */
      prnt_trt_typ ();
      prnt_ped_dat ();
      if (PedDatTyp == DISCR_LIAB)
         test_num_liab_classes ();
   }

   if (err_typ)                         /* Handle any outstanding errors.     */
      if (proc_err ())
         return FALSE;

   if (TRAIT_PENE)                      /* Checks incomplete penetrances for  */
   {                                    /* discete trait data.                */
      if (!test_penetrnc ())
         if (proc_err ())
            return FALSE;

      prnt_penetrnc ();                 /* Prints penetrances.                */
   }

   if (TRAIT_DATA)                      /* Prints a blank line following the  */
      fprintf (stdout, "\n");           /* reporting of trait pedigree data.  */

   if (n_err)
      return proc_end ();

   prnt_lods_pars ();                   /* Prints MC parameters.              */

   return TRUE;
}
