/******************************************************************************/
/*                                                                            */
/* Program: ibd_class - standalone IBDgraph main program.                     */
/* Elizabeth Thompson                                                         */
/* Lucas and Hoyt Koepke                                                      */
/* Steven Lewis                                                               */
/*                                                                            */
/******************************************************************************/

#define PNAM "phoenix"
#define DEFINE_DATA

#include "errsprogs.h"
#include "stuff.h"
#include "IBDgraph.h"

void print_help (void)
{
   printf ("\nUsage:          ./ibd_class <file> <option>\n\n");
   printf ("Options:        none            ");
   printf ("Prints equivalence classes over all markers\n");
   printf ("                -m <int>        ");
   printf ("Prints equivalent graphs at marker <>\n");
   printf ("                -r <int> <int>  ");
   printf ("Prints graphs equivalent over range <> to <>\n");
   printf ("                -s <int> <int>  ");
   printf ("Prints invariance range of graph around marker\n");
   printf ("                -a <int>        ");
   printf ("Prints entire invariance set of graph\n");
   printf ("                -e              ");
   printf ("Prints equivalence classes at resolution of the marker\n\n");

   return;
}

int main (int argc, char **argv)
{
   IBDGraphEquivalences *ibd_equivalences;
   IBDGraphLocationEquivalences *igeq;
   IBDGraphList *ibd_graphs;
   IBDGraph *g;

   HashTable *ht;
   HashTableIterator *hti;
   HashObject *h;

   markertype m, ml, mu;
   cmi_ptr mi;

   MI_LIST_NODE_PTR p_miln;

   char  help1[] = "-h";
   char  help2[] = "--help";

   long  index;

   int   marker_loc[2];
   int   r1;

   /***************************************************************************/
   /*                                                                         */
   /* The executable portion of the main program starts here.                 */
   /*                                                                         */
   /***************************************************************************/

   if (argc == 1 || strcmp (argv[1], help1) == 0 ||
       strcmp (argv[1], help2) == 0)
   {
      print_help ();
      return 0;
   }

   ibd_graphs = NewIBDGraphList ();     /* Initializing graph list.           */

   /* Handle the case where the command line consists of only a file name.    */
   if (argc == 2)
   {
      createIBDGraphs (argv[1], ibd_graphs);
      ibd_equivalences = IBDGraphEquivalenceClasses (ibd_graphs);
      IBDGraphEquivalences_InplaceSort (ibd_equivalences);

      printf ("\nEquivalence classes over all markers:\n\n");
      IBDGraphEquivalences_Print (ibd_equivalences);
      printf ("\n");
   }

   /* Parse the command line options.                                         */
   if ((argc > 2) && (argv[2][0] == '-'))
   {
      switch (argv[2][1])
      {
          case 'h':
             print_help ();
             break;

          case 'm':
             if (argc != 4)
             {
                printf
                   ("\nERROR: Incorrect number of arguments for the -m flag.\n");
                printf
                   ("Use the -h or --help flag to see options and usage.\n\n");
                exit (1);
             }

             /* Get marker location.                                          */
             marker_loc[0] = atoi (argv[3]);

             createIBDGraphs (argv[1], ibd_graphs);
             ibd_equivalences =
                IBDGraphEquivalenceClassesAtMarker (ibd_graphs, marker_loc[0]);
             IBDGraphEquivalences_InplaceSort (ibd_equivalences);

             printf ("\nTesting at specified marker location %d:\n\n",
                     marker_loc[0]);
             IBDGraphEquivalences_Print (ibd_equivalences);

             break;

          case 'r':
             if (argc != 5)
             {
                printf
                   ("\nERROR: Incorrect number of arguments for the -r flag.\n");
                printf
                   ("Use the -h or --help flag to see options and usage.\n\n");
                exit (1);
             }

             /* Get range of markers.                                         */
             marker_loc[0] = atoi (argv[3]);
             marker_loc[1] = atoi (argv[4]);

             createIBDGraphs (argv[1], ibd_graphs);
             ibd_equivalences =
                IBDGraphEquivalenceClassesOfMarkerRange (ibd_graphs,
                                                         marker_loc[0],
                                                         marker_loc[1] + 1);
             IBDGraphEquivalences_InplaceSort (ibd_equivalences);

             printf ("\nTesting range from marker %d to %d:\n\n", marker_loc[0],
                     marker_loc[1]);
             IBDGraphEquivalences_Print (ibd_equivalences);

             break;

          case 's':
             if (argc != 5)
             {
                printf
                   ("\nERROR: Incorrect number of arguments for the -s flag.\n");
                printf
                   ("Use the -h or --help flag to see options and usage.\n\n");
                exit (1);
             }

             /* Get graph number and specified marker number.                 */
             createIBDGraphs (argv[1], ibd_graphs);

             index = atoi (argv[3]) - 1;
             if (index < 0 || index >= Igl_Size (ibd_graphs))
             {
                printf
                   ("ERROR: Index %ld out of range, must be 1 <= index <= %ld.",
                    index + 1, Igl_Size (ibd_graphs));
                exit (1);
             }

             m = atol (argv[4]);
             g = Igl_ViewItem (ibd_graphs, (size_t) index);

             IBDGraphInvariantRegion (&ml, &mu, g, m);

             printf ("\nIBD graph %ld is invariant on the ", index + 1);
             printf ("interval ");
             Mi_PrintInterval (ml, mu);
             printf (".\n");

             break;

          case 'a':
             if (argc != 4)
             {
                printf
                   ("\nERROR: Incorrect number of arguments for the -a flag.\n");
                printf
                   ("Use the -h or --help flag to see options and usage.\n\n");
                exit (1);
             }

             /* Get graph number.                                             */
             createIBDGraphs (argv[1], ibd_graphs);

             index = atoi (argv[3]) - 1;
             if (index < 0 || index >= Igl_Size (ibd_graphs))
             {
                printf
                   ("ERROR: Index %ld out of range, must be 1 <= index <= %ld.",
                    index + 1, Igl_Size (ibd_graphs));
                exit (1);
             }

             g = Igl_ViewItem (ibd_graphs, (size_t) index);

             /* The next two lines produce output displaying the marker sets  */
             /* over which IBD is invariant using a debugging subroutine in   */
             /* the IBDgraph library.                                         */

             /* printf ("\nIBD graph %4ld is invariant on:\n\n", index + 1);  */
             /* IBDGraph_Print (g);                                           */

             /* The programming which follows was mostly copied from the      */
             /* IBDgraph library subroutine IBDGraph_Print.  However, in this */
             /* code the actual output of results is delayed so that the      */
             /* results are output in increasing marker number order.         */

             if (g->dirty)
                IBDGraph_Refresh (g);

             ht = g->graph_hashes;
             hti = Hti_New (ht);

             while (Hti_Next (&h, hti))
             {
                mi = H_Mi (h);
                if (mi != NULL)
                   alloc_mi_list_node ((MI_PTR) mi);
             }

             printf ("\nIBD graph %4ld is invariant on:", index + 1);

             p_miln = first_mi_list_node;
             if (p_miln == NULL)
                printf ("[-inf, inf)");

             while (p_miln)
             {
                if (p_miln != first_mi_list_node)
                   printf ("                               ");

                if (p_miln->p_mi->num_array_ranges == 0)
                   PrintInterval (p_miln->p_mi->r.start, p_miln->p_mi->r.end);
                else
                {
                   for (r1 = 0; r1 < p_miln->p_mi->num_array_ranges; ++r1)
                   {
                      PrintInterval (p_miln->p_mi->range_list[r1].start,
                                     p_miln->p_mi->range_list[r1].end);
                   }
                }

                p_miln = p_miln->p_next;
                printf ("\n");
             }

             Hti_Delete (hti);
             free_mi_list ();

             break;

          case 'e':
             if (argc != 3)
             {
                printf
                   ("\nERROR: Incorrect number of arguments for the -e flag.\n");
                printf
                   ("Use the -h or --help flag to see options and usage.\n\n");
                exit (1);
             }

             createIBDGraphs (argv[1], ibd_graphs);
             igeq = NewIBDGraphLocationEquivalences (ibd_graphs);
             IBDGraphLocationEquivalences_Print (igeq);

             printf ("\nGrouped %ld locations into %ld equivalence ",
                     IBDLocEq_TotalSize (igeq), IBDLocEq_NumClasses (igeq));
             printf ("classes.\n\n");

             O_DecRef (igeq);

             break;

          default:
             printf
                ("\nImproper argument: %s, use option -h for usage.\n\n",
                 argv[4]);
             break;
      }
   }

   O_DecRef (ibd_graphs);

   return 0;
}
