/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*  Functions are:
 *          allocPed()
 *          reallocRSDat()
 *          readPed()
 *          nextRecord()
 *          nextNam()
 *          nextInt()
 *          nextReal()
 *          checkPed()
 *          openChkPed()
 *          writChkPed()
 *          assignChkPed()
 *          freeChkPed()
 */

#include "pedchk.h"

/*----------------------------------------------------------------------------*/
/* allocPed():  Allocates arrays for storing pedigree data.                   */

void allocPed (void)
{
   pPedDat = malloc (PedSize * sizeof (PEDD));
   alloc_chk (WHERE, pPedDat);

   if (NNInts)
   {
      pIntDat = malloc (PedSize * (NNInts) * sizeof (int));
      alloc_chk (WHERE, pIntDat);
   }
   if (NReals && !DiscardReals)
   {
      maxRSDat = FLTPT_CHR * NReals * PedSize;
      ppRSDat = malloc (PedSize * sizeof (char *));
      alloc_chk (WHERE, ppRSDat);
      *ppRSDat = malloc (maxRSDat * sizeof (char));
      alloc_chk (WHERE, *ppRSDat);
   }

   return;
}

/*----------------------------------------------------------------------------*/
/* reallocRSDat(): Increases size of array for storing reals as char strings. */

void reallocRSDat (void)
{
   char *pOld;
   int   i;

   maxRSDat *= 1.2;
   pOld = *ppRSDat;
   *ppRSDat = realloc (*ppRSDat, maxRSDat * sizeof (char));
   alloc_chk (WHERE, *ppRSDat);
   /*                                 * this way to reassign pointers!
    *                                 * see 7.29, C Programming FAQ bk */
   for (i = 1; i <= nInd; ++i)
      *(ppRSDat + i) = *ppRSDat + (*(ppRSDat + i) - pOld);

   return;
}

/*---------------------------------------------------------------------------*/
/* readPed():  Reads pedigree data.  Checks data for valid format.           */
/*             Stores data in PedDat, IntDat and RSDat arrays.               */

int readPed (FILE * pF)
{
   int   i, num, nchr;
   PED_PTR pP = pPedDat;
   PEDI_PTR pI = pIntDat;
   char  pedRec[MAXLINLEN + 2];
   char *pStr, *pTok;

   int   namlen;
   widPNam = 0;

   /*                                    * Read a record for each individual. */
   while ((nInd < PedSize) && nextRecord (pedRec, pF))
   {
      pStr = pedRec;                    /* Gets individual's name, parents,
                                         * gender (if present) and
                                         * observed (if present).             */
      for (i = 0; i < 3; ++i)
      {
         if (!nextNam (&pStr, &pTok, nInd + 1))
            return OK;

         switch (StdIn[i])
         {
             case NAM:
                strncpy (pP->nam, pTok, MAXNAMLEN);
                /* terminate the string */
                pP->nam[MAXNAMLEN] = '\0';
                /* figure out the max length of the names */
                namlen = strlen (pP->nam);
                if (widPNam < namlen)
                   widPNam = namlen;
                break;
             case PA:
                strncpy (pP->namPa, pTok, MAXNAMLEN);
                pP->namPa[MAXNAMLEN] = '\0';
                break;
             case MA:
                strncpy (pP->namMa, pTok, MAXNAMLEN);
                pP->namMa[MAXNAMLEN] = '\0';
                break;
             default:
                break;
         }
         pStr = NULL;                   /* Required for strtok().             */
      }

      for (i = 3; i < NStdIn; ++i)
      {
         if (!nextInt (&pStr, &pTok, &num, nInd + 1))
            return !OK;
         if (num < 0)
         {
            give_err (ERC_fin, pTok, nInd + 1);
            return !OK;
         }
         switch (StdIn[i])
         {
             case SEX:
                {
                   pP->sex = num;
                   if (pP->sex != MALE && pP->sex != FEMALE
                       && pP->sex != UNSEXED)
                   {
                      give_err (ERC_fsx, pTok, nInd + 1);
                      return !OK;
                   }
                   break;
                }
             case OBS:
                {
                   pP->obsrvd = num;
                   break;
                }
             default:
                break;
         }
         pStr = NULL;                   /* Required for strtok().             */
      }
      /*                                 * Tests for option to assign
       *                                 * or ignore gender                 */
      if (SexOp == ASSIGN)
         pP->sex = UNSEXED;
      /*                                 * Tests for option to observe      */
      if (ObsOp == ALL)                 /* all individuals                  */
         pP->obsrvd = OBSERVED;

      /*                                 * gets the rest of the integers    */
      for (i = 0; i < NNInts; ++i)
      {
         if (!nextInt (&pStr, &pTok, &num, nInd + 1))
            return !OK;

         *(pI + i) = num;
         if (num > maxPInt[i])
            maxPInt[i] = num;
         else if (num < minPInt[i])
            minPInt[i] = num;
      }
      /*                                 * checks the reals                 */
      if (NReals && !DiscardReals)      /* and stores the tokens            */
      {
         *(ppRSDat + nInd) = *ppRSDat + curRSDat;

         for (i = 0; i < NReals; ++i)
         {
            if (!nextReal (&pStr, &pTok, nInd + 1))
               return !OK;
            nchr = strlen (pTok);

            while (curRSDat + nchr + 3 >= maxRSDat)
               reallocRSDat ();

            strcpy (*ppRSDat + curRSDat, "  ");
            curRSDat += 2;
            strcpy (*ppRSDat + curRSDat, pTok);
            curRSDat += nchr;
         }
         strcpy (*ppRSDat + curRSDat, "\0");
         ++curRSDat;
      }
      ++pP;
      pI += NNInts;
      ++nInd;
   }
   return OK;
}

/******************************************************************************/
/*                                                                            */
/* The nextRecord subroutine can be called to advance to the next pedigree    */
/* record containing something other than WHITESPACE.                         */
/*                                                                            */
/******************************************************************************/

int nextRecord (char *pedRec, FILE * pF)
{
   char *p_rec;                         /* Pointer to next pedigree record.   */
   size_t l_rec;                        /* Length of the pedigree record.     */
   size_t l_ws;                         /* Length of the WHITESPACE.          */

   /* Keep reading as many records from the pedigree file as is necessary to  */
   /* find a record which contains something other than WHITESPACE in it.     */

   p_rec = NULL;

   while (p_rec == NULL)
   {
      p_rec = fgets (pedRec, MAXLINLEN + 1, pF);
      if (p_rec == NULL)
         return !OK;                    /* Tell the caller that EOF reached.  */

      l_rec = strlen (p_rec);           /* Find out how long the record is.   */

      /* Find the longest span of WHITESPACE in the pedigree record.          */
      l_ws = strspn (p_rec, WHITESPACE);

      /* If this equals the record length, then this pedigree record consists */
      /* entirely of WHITESPACE.                                              */
      if (l_ws == l_rec)
         p_rec = NULL;
   }

   /* Otherwise the pedigree record contains something other than WHITESPACE. */
   return OK;
}

int nextNam (char **ppS, char **ppT, int nR)
{
   if (!(*ppT = strtok (*ppS, WHITESPACE)))
   {
      give_err (ERC_frm, nR);
      return !OK;
   }

   return OK;
}

int nextInt (char **ppS, char **ppT, int *n, int nR)
{                                       /* Converts character to integer.     */
   char *pEnd;

   if (!(*ppT = strtok (*ppS, WHITESPACE)))
   {
      give_err (ERC_frm, nR);
      return !OK;
   }

   errno = 0;
   *n = strtol (*ppT, &pEnd, 10);

   if (*pEnd != NUL)
   {
      give_err (ERC_fbi, *ppT, nR);
      return !OK;
   }

   if (errno == ERANGE)
   {
      give_err (ERC_fri, *ppT, nR);
      return !OK;
   }

   return OK;
}

/******************************************************************************/
/*                                                                            */
/* The nextReal subroutine checks that the next token available in the input  */
/* stream pointed to by the first argument can be interpretted as a valid     */
/* double precision floating point number.  If it is a valid double precision */
/* floating point number, the second argument is set to point at the token    */
/* and the subroutine exits with a return code reflecting that the token is   */
/* OK.  If it is not a valid double precision floating point number the       */
/* subroutine generates an error message using the third argument to indicate */
/* for which individual there was an invalid floating point number read in    */
/* from the pedigree file and then exits with a return code reflecting that   */
/* the token is not OK.                                                       */
/*                                                                            */
/******************************************************************************/

int nextReal (char **ppS, char **ppT, int nR)
{
   char *pEnd;

   if (!(*ppT = strtok (*ppS, WHITESPACE)))
   {
      give_err (ERC_frm, nR);
      return !OK;
   }

   errno = 0;
   (void) strtod (*ppT, &pEnd);

   if (*pEnd != NUL)
   {
      give_err (ERC_fbr, *ppT, nR);
      return !OK;
   }

   if (errno == ERANGE)
   {
      give_err (ERC_frr, *ppT, nR);
      return !OK;
   }

   return OK;
}

/*---------------------------------------------------------------------------*/
/* checkPed(): Checks pedigree size.                                         */

int checkPed (FILE * pF)
{
   int   r;

   if (PrintLevel >= 3)
   {
      fprintf (stdout, " %d individuals read\n\n", nInd);

      if (ObsOp == ALL)
         fprintf (stdout, " All individuals are observed\n\n");
   }

   if (nInd < PedSize)
      give_warn (ERR_nex, PedSize);

   if (nInd < 3)
   {
      give_err (ERR_siz, "checkPed");
      return !OK;
   }
   else
   {
      r = fscanf (pF, "%*s");
      if (r != EOF)
         give_warn (ERR_urd, "checkPed");
   }

   return OK;
}

/*---------------------------------------------------------------------------*/
/* openChkPed(): Opens the output pedigree file.                             */

FILE *openChkPed (void)
{
   int   i, found;
   FILE *pOF;

   if (!get_filent (OPED)->fname)
   {
      give_err (ERC_nof);
      return !OK;
   }
   if (!(pOF = open_fil (OPED)))
      return NULL;

   fprintf (pOF, "input pedigree size %d\n", nInd);

   fprintf (pOF, "input pedigree record names 3");
   if (NStdOut - 3 + NNInts > 0)
      fprintf (pOF, " integers %d", NStdOut - 3 + NNInts);
   if (NReals && !DiscardReals)
      fprintf (pOF, " reals %d\n", NReals);
   fprintf (pOF, "\n");

   if (StdOut[1] == PA)
      fprintf (pOF, "input pedigree record father mother\n");
   else
      fprintf (pOF, "input pedigree record mother father\n");

   found = N;
   for (i = 0; i < NStdOut; ++i)
      if (StdOut[i] == SEX)
         found = Y;
   if (found == Y)
      fprintf (pOF, "input pedigree record gender present\n");
   else
      fprintf (pOF, "input pedigree record gender absent\n");

   found = N;
   for (i = 0; i < NStdOut; ++i)
      if (StdOut[i] == OBS)
         found = Y;
   if (found == Y)
      fprintf (pOF, "input pedigree record observed present\n");
   else
      fprintf (pOF, "input pedigree record observed absent\n");

   fprintf (pOF, "********\n");

   return pOF;
}

/*---------------------------------------------------------------------------*/
/* writChkPed():  Writes the pedigree file, in chronological order           */
/*                contains the same data as the input file, except reals     */
/*                may be discarded.                                          */

void writChkPed (FILE * pOF)
{
   KEYS_PP ppK = ppKeyTab;
   PED_PTR pP;
   PEDI_PTR pPI;

   int   i, j, ixInd;

   /* gets field widths for format   */
   for (j = 0; j < NNInts; ++j)
      widPInt[j] = intWid (minPInt[j], maxPInt[j]);

   /*for each individual:  */
   for (i = 0; i < nInd; ++i)
   {
      ixInd = (*ppK)->ix - 1;           /* gets index for next individual */

      pP = pPedDat + ixInd;             /* outputs standard ints
                                         * (name, parents, gender, & if
                                         * required, observed indicator)  */
      for (j = 0; j < NStdOut; ++j)
      {
         switch (StdOut[j])
         {
             case NAM:
                {
                   fprintf (pOF, "%*s", widPNam, pP->nam);
                   break;
                }
             case PA:
                {
                   fprintf (pOF, " %*s", widPNam, pP->namPa);
                   break;
                }
             case MA:
                {
                   fprintf (pOF, " %*s", widPNam, pP->namMa);
                   break;
                }
             case SEX:
                {
                   fprintf (pOF, "  %d", pP->sex);
                   break;
                }
             case OBS:
                {
                   fprintf (pOF, "  %d", pP->obsrvd);
                   break;
                }
             default:
                break;
         }
      }
      fprintf (pOF, " ");               /* outputs integers               */

      pPI = pIntDat + NNInts * ixInd;

      for (j = 0; j < NNInts; ++j)
         fprintf (pOF, " %*d", widPInt[j], *(pPI + j));

      if (NReals && !DiscardReals)      /* outputs reals                  */
         fprintf (pOF, "%s", *(ppRSDat + ixInd));

      fprintf (pOF, "\n");              /* ends current record            */
      fflush (pOF);

      ++ppK;
   }
   PedSize = nInd;                      /* set PedSize to nInd            */

   return;
}

/*---------------------------------------------------------------------------*/
/* assignChkPed():  Frees the name of the input pedigree file                */
/*                  and resets its name to the name of the output file       */
/*                  sets a NULL pointer to the name of the output file       */
/*                  resets globals for input file to that of the output      */
/*                  file.                                                    */

int assignChkPed (void)
{
   int   i;

   free_fnam (get_filent (PED));

   if (!(get_filent (PED)->fname = get_filent (OPED)->fname))
   {
      give_err (ERC_nof);
      return !OK;
   }
   get_filent (OPED)->fname = NULL;

   NStdIn = NStdOut;
   NInts = NNInts + NStdOut - 3;

   for (i = 0; i < NStdOut; ++i)
      StdIn[i] = StdOut[i];

   SexIn = SexOut;
   ObsIn = ObsOut;

   return OK;
}

/*---------------------------------------------------------------------------*/
/* freeChkPed(): Frees pedigree data arrays.                                 */

void freeChkPed (void)
{
   if (pPedDat)
   {
      free (pPedDat);
      pPedDat = NULL;
   }
   if (pIntDat)
   {
      free (pIntDat);
      pIntDat = NULL;
   }
   if (ppRSDat)
   {
      free (*ppRSDat);
      *ppRSDat = NULL;
      free (ppRSDat);
      ppRSDat = NULL;
   }

   return;
}
