/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/* Includes the following MORGAN file handling functions:

 *        open_chk ()
 *        open_chkq ()
 *        open_fil ()
 *        open_ped ()
 *        open_ind ()
 *        get_filent ()
 *        get_filtyp ()
 *        free_fnam ()
 *        free_fnams ()
 */

#include "stuff.h"

/* --------------- open_chk: Opens file with the specified mode.  --------------
 *                           Prints a message for the file opened.
 *                           Prints an error message and quits if unrecognized
 *                           file mode or if unable to open file.
 */

FILE *open_chk (char *src_fil, int lin_no, char *fnam, char *mode)
{
   FILE *pf = NULL;
   int   i;
   struct _fil_modes
   {
      char *typ;
      char *txt;
   };
   struct _fil_modes fmodes[] = {
      {"r", "reading"},
      {"w", "writing"},
      {"a", "appending"},
      {"r+", "update"},
      {"w+", "update"},
      {"a+", "update"},
      {"rb", "reading"},
      {"wb", "writing"},
      {"ab", "appending"},
      {"r+b", "update"},
      {"w+b", "update"},
      {"a+b", "update"},
      {"rb+", "update"},
      {"wb+", "update"},
      {"ab+", "update"},
      {"", "unrecognized mode"}
   };

   for (i = 0; i < 15; ++i)
      if (!strcmp (mode, fmodes[i].typ))
         break;

   if ((pf = fopen (fnam, mode)))
   {
      if (i < 6)
         fprintf (stdout, " Opened file \"%s\" for ", fnam);
      else if (i < 15)
         fprintf (stdout, " Opened binary file \"%s\" for ", fnam);
      else
         fprintf (stdout, " Opened file \"%s\" for ", fnam);
      fprintf (stdout, "%s\n\n", fmodes[i].txt);
   }
   else
      give_up_where (src_fil, lin_no, ERS_fop, fnam, mode, strerror (errno));

   return pf;
}

/* -------------- open_chkq: Opens file with the specified mode.  --------------
 *                           Like open_chk(), but "quiet"
 *                           (does not report opening of file).
 */

FILE *open_chkq (char *src_fil, int lin_no, char *fnam, char *mode)
{
   FILE *pf = NULL;

   if (!(pf = fopen (fnam, mode)))
      give_up_where (src_fil, lin_no, ERS_fop, fnam, mode, strerror (errno));

   return pf;
}

/* Opens a Morgan file for reading or writing.
 * Prints an error message and returns NULL if "write only" file already exists,
 * or if unable to open file.                  MAJ Aug 1999, 2005.
 */

FILE *open_fil (char *fmnem)
/* Argument is one of the file type mnemonics defined for Filtab[].           */
/* Returns a FILE pointer.                                                    */
{
   FILTAB_PTR ptb;
   FILE *pf;

   ptb = get_filent (fmnem);

   /* Checks for output file already existing.                                */
   if (ptb->fname && strcmp (ptb->fname, DATASINK)) /* (Unless data sink).    */
   {
      /* Checks for file to be overwritten.                                   */
      if (!strcmp (ptb->fop, "w"))
         if ((pf = fopen (ptb->fname, "r")))
         {
            if (ptb->rewrit == Y)
            {
               fclose (pf);
               give_warn (ERS_fxr, toupper ((int) *ptb->fdesc),
                          ptb->fdesc + 1, ptb->fname);
            }
            else
            {
               fclose (pf);
               give_err (ERS_fex, toupper ((int) *ptb->fdesc),
                         ptb->fdesc + 1, ptb->fname);
               return NULL;
            }
         }

      /* Checks for file to be overwritten rather than appended.              */
      if (!strcmp (ptb->fop, "a") && ptb->rewrit == Y)
         if ((pf = fopen (ptb->fname, "r")))
         {
            fclose (pf);
            give_warn (ERS_fxr, toupper ((int) *ptb->fdesc),
                       ptb->fdesc + 1, ptb->fname);
            if ((pf = fopen (ptb->fname, "w")))
               fclose (pf);
         }
   }

   /* Opens file.                                                             */
   if ((pf = fopen (ptb->fname, ptb->fop)))
      fprintf (stdout, " Opened %s \"%s\"\n\n", ptb->fdesc, ptb->fname);
   else
      give_err (ERS_fof, ptb->fdesc, ptb->fname, strerror (errno));

   return pf;
}

/* --------------- open_ped: Called by MORGAN programs.  -----------------------
 *                           Opens or reopens the pedigree file, skipping to
 *                           the end-of-input-parameters indicator, "****".
 *                           WARNING: gives up if not successful.
 */

FILE *open_ped (void)
{
   FILE *pf;
   char  nxt_lin[MAXLINLEN + 2];
   char *pstr, *ptok;

   if ((pf = open_fil (PED)))
   {
      while ((pstr = fgets (nxt_lin, MAXLINLEN + 1, pf)) &&
             (!(ptok = strtok (pstr, WHITESPACE)) || strncmp (ENDPAR, ptok, 4)))
         ;                              /* Continue.                          */

      if (pstr == NULL)
      {
         pf = NULL;
         give_err (ERS_pfe, ENDPAR);
      }
   }

   return pf;
}

/* --------------- open_ind: Called by MORGAN programs.  -----------------------
 *                           Opens or reopens the individuals file, skipping to
 *                           the end-of-input-parameters indicator, "****".
 *                           WARNING: gives up if not successful.
 */

FILE *open_ind (void)
{
   FILE *pf;
   char  nxt_lin[MAXLINLEN + 2];
   char *pstr, *ptok;

   if ((pf = open_fil (IND)))
   {
      while ((pstr = fgets (nxt_lin, MAXLINLEN + 1, pf)) &&
             (!(ptok = strtok (pstr, WHITESPACE)) || strncmp (ENDPAR, ptok, 4)))
         ;                              /* Continue.                          */

      if (pstr == NULL)
      {
         pf = NULL;
         give_err (ERS_pfe, ENDPAR);
      }
   }

   return pf;
}

/* ------------- get_filent: Called by MORGAN functions.  ----------------------
 *                           Finds the Filtab[] entry for specified file type.
 *                              MAJ Aug 1999.
 */

FILTAB_PTR get_filent (char *f_mnem)    /* Argument is one of the file type   */
                                        /* mnemonics defined for Filtab[].    */
                                        /* Returns a pointer to Filtab[]      */
                                        /* or NULL pointer.                   */
{
   FILTAB_PTR ptb;

   for (ptb = Filtab; ptb->type[0] != NUL
        && strncmp (ptb->type, f_mnem, 3); ++ptb)
      ;                                 /* Continue.                          */

   return (ptb->type[0] == NUL) ? NULL : ptb;
}

/* ------------- get_filtyp: Called by MORGAN functions.  ----------------------
 *                           Finds the next file type permitted for the
 *                           specified program.
 */

FILTAB_PTR get_filtyp (char prog)       /* Argument is one of the single      */
                                        /* character program codes in         */
                                        /* Filtab[].                          */
                                        /* Returns a pointer to next          */
                                        /* Filtab[] entry for program         */
                                        /* or NULL pointer.                   */
{
   static FILTAB_PTR ptb;

   if (!ptb)
      ptb = Filtab + 1;
   else
      ++ptb;
   while (ptb && ptb->type[0] != NUL && !(strchr (ptb->pcods, prog)))
      ++ptb;

   return (ptb->type[0] == NUL) ? NULL : ptb;
}

/* -------------- free_fnam: Called by MORGAN programs.  -----------------------
 *                           Frees the string containing the file name.
 */

void free_fnam (FILTAB_PTR pftab)
{
   if (pftab && pftab->fname)
   {
      free ((void *) pftab->fname);
      pftab->fname = NULL;
   }

   return;
}

/* ------------- free_fnams: Called by MORGAN programs.  -----------------------
 *                           Cleanup function.
 *                           Frees file names referenced in Filtab[].
 */

void free_fnams (void)
{
   FILTAB_PTR ptb;

   for (ptb = Filtab; ptb->type[0] != NUL; ++ptb)
      free_fnam (ptb);

   return;
}
